from diofant.concrete import Sum
from diofant.concrete.delta import deltaproduct as dp
from diofant.concrete.delta import deltasummation as ds
from diofant.core import Eq, Integer, oo, symbols
from diofant.functions import KroneckerDelta as Kd
from diofant.functions import Piecewise, exp, piecewise_fold
from diofant.logic import And


__all__ = ()

i, j, k, l, m = symbols("i j k l m", integer=True, finite=True)
x, y = symbols("x y", commutative=False)


def test_deltaproduct_trivial():
    assert dp(x, (j, 1, 0)) == 1
    assert dp(x, (j, 1, 3)) == x**3
    assert dp(x + y, (j, 1, 3)) == (x + y)**3
    assert dp(x*y, (j, 1, 3)) == (x*y)**3
    assert dp(Kd(i, j), (k, 1, 3)) == Kd(i, j)
    assert dp(x*Kd(i, j), (k, 1, 3)) == x**3*Kd(i, j)
    assert dp(x*y*Kd(i, j), (k, 1, 3)) == (x*y)**3*Kd(i, j)


def test_deltaproduct_basic():
    assert dp(Kd(i, j), (j, 1, 3)) == 0
    assert dp(Kd(i, j), (j, 1, 1)) == Kd(i, 1)
    assert dp(Kd(i, j), (j, 2, 2)) == Kd(i, 2)
    assert dp(Kd(i, j), (j, 3, 3)) == Kd(i, 3)
    assert dp(Kd(i, j), (j, 1, k)) == Kd(i, 1)*Kd(k, 1) + Kd(k, 0)
    assert dp(Kd(i, j), (j, k, 3)) == Kd(i, 3)*Kd(k, 3) + Kd(k, 4)
    assert dp(Kd(i, j), (j, k, l)) == Kd(i, l)*Kd(k, l) + Kd(k, l + 1)
    assert dp(Kd(i, 1), (i, j**2, k**2)) == (Kd(1, j**2)*Kd(j**2, k**2) +
                                             Kd(k**2, j**2 - 1))


def test_deltaproduct_mul_x_kd():
    assert dp(x*Kd(i, j), (j, 1, 3)) == 0
    assert dp(x*Kd(i, j), (j, 1, 1)) == x*Kd(i, 1)
    assert dp(x*Kd(i, j), (j, 2, 2)) == x*Kd(i, 2)
    assert dp(x*Kd(i, j), (j, 3, 3)) == x*Kd(i, 3)
    assert dp(x*Kd(i, j), (j, 1, k)) == x*Kd(i, 1)*Kd(k, 1) + Kd(k, 0)
    assert dp(x*Kd(i, j), (j, k, 3)) == x*Kd(i, 3)*Kd(k, 3) + Kd(k, 4)
    assert dp(x*Kd(i, j), (j, k, l)) == x*Kd(i, l)*Kd(k, l) + Kd(k, l + 1)


def test_deltaproduct_mul_add_x_y_kd():
    assert dp((x + y)*Kd(i, j), (j, 1, 3)) == 0
    assert dp((x + y)*Kd(i, j), (j, 1, 1)) == (x + y)*Kd(i, 1)
    assert dp((x + y)*Kd(i, j), (j, 2, 2)) == (x + y)*Kd(i, 2)
    assert dp((x + y)*Kd(i, j), (j, 3, 3)) == (x + y)*Kd(i, 3)
    assert dp((x + y)*Kd(i, j), (j, 1, k)) == \
        (x + y)*Kd(i, 1)*Kd(k, 1) + Kd(k, 0)
    assert dp((x + y)*Kd(i, j), (j, k, 3)) == \
        (x + y)*Kd(i, 3)*Kd(k, 3) + Kd(k, 4)
    assert dp((x + y)*Kd(i, j), (j, k, l)) == \
        (x + y)*Kd(i, l)*Kd(k, l) + Kd(k, l + 1)


def test_deltaproduct_add_kd_kd():
    assert dp(Kd(i, k) + Kd(j, k), (k, 1, 3)) == 0
    assert dp(Kd(i, k) + Kd(j, k), (k, 1, 1)) == Kd(i, 1) + Kd(j, 1)
    assert dp(Kd(i, k) + Kd(j, k), (k, 2, 2)) == Kd(i, 2) + Kd(j, 2)
    assert dp(Kd(i, k) + Kd(j, k), (k, 3, 3)) == Kd(i, 3) + Kd(j, 3)
    assert dp(Kd(i, k) + Kd(j, k), (k, 1, l)) == Kd(l, 0) + \
        Kd(i, 1)*Kd(l, 1) + Kd(j, 1)*Kd(l, 1) + \
        Kd(i, 1)*Kd(j, 2)*Kd(l, 2) + Kd(j, 1)*Kd(i, 2)*Kd(l, 2)
    assert dp(Kd(i, k) + Kd(j, k), (k, l, 3)) == Kd(l, 4) + \
        Kd(i, 3)*Kd(l, 3) + Kd(j, 3)*Kd(l, 3) + \
        Kd(i, 2)*Kd(j, 3)*Kd(l, 2) + Kd(i, 3)*Kd(j, 2)*Kd(l, 2)
    assert dp(Kd(i, k) + Kd(j, k), (k, l, m)) == Kd(l, m + 1) + \
        Kd(i, m)*Kd(l, m) + Kd(j, m)*Kd(l, m) + \
        Kd(i, m)*Kd(j, m - 1)*Kd(l, m - 1) + Kd(i, m - 1)*Kd(j, m)*Kd(l, m - 1)


def test_deltaproduct_mul_x_add_kd_kd():
    assert dp(x*(Kd(i, k) + Kd(j, k)), (k, 1, 3)) == 0
    assert dp(x*(Kd(i, k) + Kd(j, k)), (k, 1, 1)) == x*(Kd(i, 1) + Kd(j, 1))
    assert dp(x*(Kd(i, k) + Kd(j, k)), (k, 2, 2)) == x*(Kd(i, 2) + Kd(j, 2))
    assert dp(x*(Kd(i, k) + Kd(j, k)), (k, 3, 3)) == x*(Kd(i, 3) + Kd(j, 3))
    assert dp(x*(Kd(i, k) + Kd(j, k)), (k, 1, l)) == Kd(l, 0) + \
        x*Kd(i, 1)*Kd(l, 1) + x*Kd(j, 1)*Kd(l, 1) + \
        x**2*Kd(i, 1)*Kd(j, 2)*Kd(l, 2) + x**2*Kd(j, 1)*Kd(i, 2)*Kd(l, 2)
    assert dp(x*(Kd(i, k) + Kd(j, k)), (k, l, 3)) == Kd(l, 4) + \
        x*Kd(i, 3)*Kd(l, 3) + x*Kd(j, 3)*Kd(l, 3) + \
        x**2*Kd(i, 2)*Kd(j, 3)*Kd(l, 2) + x**2*Kd(i, 3)*Kd(j, 2)*Kd(l, 2)
    assert dp(x*(Kd(i, k) + Kd(j, k)), (k, l, m)) == Kd(l, m + 1) + \
        x*Kd(i, m)*Kd(l, m) + x*Kd(j, m)*Kd(l, m) + \
        x**2*Kd(i, m - 1)*Kd(j, m)*Kd(l, m - 1) + \
        x**2*Kd(i, m)*Kd(j, m - 1)*Kd(l, m - 1)


def test_deltaproduct_mul_add_x_y_add_kd_kd():
    assert dp((x + y)*(Kd(i, k) + Kd(j, k)), (k, 1, 3)) == 0
    assert dp((x + y)*(Kd(i, k) + Kd(j, k)), (k, 1, 1)) == \
        (x + y)*(Kd(i, 1) + Kd(j, 1))
    assert dp((x + y)*(Kd(i, k) + Kd(j, k)), (k, 2, 2)) == \
        (x + y)*(Kd(i, 2) + Kd(j, 2))
    assert dp((x + y)*(Kd(i, k) + Kd(j, k)), (k, 3, 3)) == \
        (x + y)*(Kd(i, 3) + Kd(j, 3))
    assert dp((x + y)*(Kd(i, k) + Kd(j, k)), (k, 1, l)) == Kd(l, 0) + \
        (x + y)*Kd(i, 1)*Kd(l, 1) + (x + y)*Kd(j, 1)*Kd(l, 1) + \
        (x + y)**2*Kd(i, 1)*Kd(j, 2)*Kd(l, 2) + \
        (x + y)**2*Kd(j, 1)*Kd(i, 2)*Kd(l, 2)
    assert dp((x + y)*(Kd(i, k) + Kd(j, k)), (k, l, 3)) == Kd(l, 4) + \
        (x + y)*Kd(i, 3)*Kd(l, 3) + (x + y)*Kd(j, 3)*Kd(l, 3) + \
        (x + y)**2*Kd(i, 2)*Kd(j, 3)*Kd(l, 2) + \
        (x + y)**2*Kd(i, 3)*Kd(j, 2)*Kd(l, 2)
    assert dp((x + y)*(Kd(i, k) + Kd(j, k)), (k, l, m)) == Kd(l, m + 1) + \
        (x + y)*Kd(i, m)*Kd(l, m) + (x + y)*Kd(j, m)*Kd(l, m) + \
        (x + y)**2*Kd(i, m - 1)*Kd(j, m)*Kd(l, m - 1) + \
        (x + y)**2*Kd(i, m)*Kd(j, m - 1)*Kd(l, m - 1)


def test_deltaproduct_add_mul_x_y_mul_x_kd():
    assert dp(x*y + x*Kd(i, j), (j, 1, 3)) == (x*y)**3 + \
        x*(x*y)**2*Kd(i, 1) + (x*y)*x*(x*y)*Kd(i, 2) + (x*y)**2*x*Kd(i, 3)
    assert dp(x*y + x*Kd(i, j), (j, 1, 1)) == x*y + x*Kd(i, 1)
    assert dp(x*y + x*Kd(i, j), (j, 2, 2)) == x*y + x*Kd(i, 2)
    assert dp(x*y + x*Kd(i, j), (j, 3, 3)) == x*y + x*Kd(i, 3)
    assert dp(x*y + x*Kd(i, j), (j, 1, k)) == \
        (x*y)**k + Piecewise(
            ((x*y)**(i - 1)*x*(x*y)**(k - i), And(Integer(1) <= i, i <= k)),
            (0, True)
    )
    assert dp(x*y + x*Kd(i, j), (j, k, 3)) == \
        (x*y)**(-k + 4) + Piecewise(
            ((x*y)**(i - k)*x*(x*y)**(3 - i), And(k <= i, i <= 3)),
            (0, True)
    )
    assert dp(x*y + x*Kd(i, j), (j, k, l)) == \
        (x*y)**(-k + l + 1) + Piecewise(
            ((x*y)**(i - k)*x*(x*y)**(l - i), And(k <= i, i <= l)),
            (0, True)
    )


def test_deltaproduct_mul_x_add_y_kd():
    assert dp(x*(y + Kd(i, j)), (j, 1, 3)) == (x*y)**3 + \
        x*(x*y)**2*Kd(i, 1) + (x*y)*x*(x*y)*Kd(i, 2) + (x*y)**2*x*Kd(i, 3)
    assert dp(x*(y + Kd(i, j)), (j, 1, 1)) == x*(y + Kd(i, 1))
    assert dp(x*(y + Kd(i, j)), (j, 2, 2)) == x*(y + Kd(i, 2))
    assert dp(x*(y + Kd(i, j)), (j, 3, 3)) == x*(y + Kd(i, 3))
    assert dp(x*(y + Kd(i, j)), (j, 1, k)) == \
        (x*y)**k + Piecewise(
            ((x*y)**(i - 1)*x*(x*y)**(k - i), And(Integer(1) <= i, i <= k)),
            (0, True)
    )
    assert dp(x*(y + Kd(i, j)), (j, k, 3)) == \
        (x*y)**(-k + 4) + Piecewise(
            ((x*y)**(i - k)*x*(x*y)**(3 - i), And(k <= i, i <= 3)),
            (0, True)
    )
    assert dp(x*(y + Kd(i, j)), (j, k, l)) == \
        (x*y)**(-k + l + 1) + Piecewise(
            ((x*y)**(i - k)*x*(x*y)**(l - i), And(k <= i, i <= l)),
            (0, True)
    )


def test_deltaproduct_mul_x_add_y_twokd():
    assert dp(x*(y + 2*Kd(i, j)), (j, 1, 3)) == (x*y)**3 + \
        2*x*(x*y)**2*Kd(i, 1) + 2*x*y*x*x*y*Kd(i, 2) + 2*(x*y)**2*x*Kd(i, 3)
    assert dp(x*(y + 2*Kd(i, j)), (j, 1, 1)) == x*(y + 2*Kd(i, 1))
    assert dp(x*(y + 2*Kd(i, j)), (j, 2, 2)) == x*(y + 2*Kd(i, 2))
    assert dp(x*(y + 2*Kd(i, j)), (j, 3, 3)) == x*(y + 2*Kd(i, 3))
    assert dp(x*(y + 2*Kd(i, j)), (j, 1, k)) == \
        (x*y)**k + Piecewise(
            (2*(x*y)**(i - 1)*x*(x*y)**(k - i), And(Integer(1) <= i, i <= k)),
            (0, True)
    )
    assert dp(x*(y + 2*Kd(i, j)), (j, k, 3)) == \
        (x*y)**(-k + 4) + Piecewise(
            (2*(x*y)**(i - k)*x*(x*y)**(3 - i), And(k <= i, i <= 3)),
            (0, True)
    )
    assert dp(x*(y + 2*Kd(i, j)), (j, k, l)) == \
        (x*y)**(-k + l + 1) + Piecewise(
            (2*(x*y)**(i - k)*x*(x*y)**(l - i), And(k <= i, i <= l)),
            (0, True)
    )


def test_deltaproduct_mul_add_x_y_add_y_kd():
    assert dp((x + y)*(y + Kd(i, j)), (j, 1, 3)) == ((x + y)*y)**3 + \
        (x + y)*((x + y)*y)**2*Kd(i, 1) + \
        (x + y)*y*(x + y)**2*y*Kd(i, 2) + \
        ((x + y)*y)**2*(x + y)*Kd(i, 3)
    assert dp((x + y)*(y + Kd(i, j)), (j, 1, 1)) == (x + y)*(y + Kd(i, 1))
    assert dp((x + y)*(y + Kd(i, j)), (j, 2, 2)) == (x + y)*(y + Kd(i, 2))
    assert dp((x + y)*(y + Kd(i, j)), (j, 3, 3)) == (x + y)*(y + Kd(i, 3))
    assert dp((x + y)*(y + Kd(i, j)), (j, 1, k)) == \
        ((x + y)*y)**k + Piecewise(
            (((x + y)*y)**(i - 1)*(x + y)*((x + y)*y)**(k - i),
             And(Integer(1) <= i, i <= k)),
            (0, True)
    )
    assert dp((x + y)*(y + Kd(i, j)), (j, k, 3)) == \
        ((x + y)*y)**(-k + 4) + Piecewise(
            (((x + y)*y)**(i - k)*(x + y)*((x + y)*y)**(3 - i),
             And(k <= i, i <= 3)),
            (0, True)
    )
    assert dp((x + y)*(y + Kd(i, j)), (j, k, l)) == \
        ((x + y)*y)**(-k + l + 1) + Piecewise(
            (((x + y)*y)**(i - k)*(x + y)*((x + y)*y)**(l - i),
             And(k <= i, i <= l)),
            (0, True)
    )


def test_deltaproduct_mul_add_x_kd_add_y_kd():
    assert dp((x + Kd(i, k))*(y + Kd(i, j)), (j, 1, 3)) == \
        Kd(i, 1)*(Kd(i, k) + x)*((Kd(i, k) + x)*y)**2 + \
        Kd(i, 2)*(Kd(i, k) + x)*y*(Kd(i, k) + x)**2*y + \
        Kd(i, 3)*((Kd(i, k) + x)*y)**2*(Kd(i, k) + x) + \
        ((Kd(i, k) + x)*y)**3
    assert dp((x + Kd(i, k))*(y + Kd(i, j)), (j, 1, 1)) == \
        (x + Kd(i, k))*(y + Kd(i, 1))
    assert dp((x + Kd(i, k))*(y + Kd(i, j)), (j, 2, 2)) == \
        (x + Kd(i, k))*(y + Kd(i, 2))
    assert dp((x + Kd(i, k))*(y + Kd(i, j)), (j, 3, 3)) == \
        (x + Kd(i, k))*(y + Kd(i, 3))
    assert dp((x + Kd(i, k))*(y + Kd(i, j)), (j, 1, k)) == \
        ((x + Kd(i, k))*y)**k + Piecewise(
            (((x + Kd(i, k))*y)**(i - 1)*(x + Kd(i, k)) *
             ((x + Kd(i, k))*y)**(-i + k), And(Integer(1) <= i, i <= k)),
            (0, True)
    )
    assert dp((x + Kd(i, k))*(y + Kd(i, j)), (j, k, 3)) == \
        ((x + Kd(i, k))*y)**(4 - k) + Piecewise(
            (((x + Kd(i, k))*y)**(i - k)*(x + Kd(i, k)) *
             ((x + Kd(i, k))*y)**(-i + 3), And(k <= i, i <= 3)),
            (0, True)
    )
    assert dp((x + Kd(i, k))*(y + Kd(i, j)), (j, k, l)) == \
        ((x + Kd(i, k))*y)**(-k + l + 1) + Piecewise(
            (((x + Kd(i, k))*y)**(i - k)*(x + Kd(i, k)) *
             ((x + Kd(i, k))*y)**(-i + l), And(k <= i, i <= l)),
            (0, True)
    )


def test_deltasummation_trivial():
    assert ds(x, (j, 1, 0)) == 0
    assert ds(x, (j, 1, 3)) == 3*x
    assert ds(x + y, (j, 1, 3)) == 3*(x + y)
    assert ds(x*y, (j, 1, 3)) == 3*x*y
    assert ds(Kd(i, j), (k, 1, 3)) == 3*Kd(i, j)
    assert ds(x*Kd(i, j), (k, 1, 3)) == 3*x*Kd(i, j)
    assert ds(x*y*Kd(i, j), (k, 1, 3)) == 3*x*y*Kd(i, j)


def test_deltasummation_basic_numerical():
    n = symbols('n', integer=True, nonzero=True)
    assert ds(Kd(n, 0), (n, 1, 3)) == 0

    # return unevaluated, until it gets implemented
    assert ds(Kd(i**2, j**2), (j, -oo, oo)) == \
        Sum(Kd(i**2, j**2), (j, -oo, oo))

    assert Piecewise((Kd(i, k), And(Integer(1) <= i, i <= 3)), (0, True)) == \
        ds(Kd(i, j)*Kd(j, k), (j, 1, 3)) == \
        ds(Kd(j, k)*Kd(i, j), (j, 1, 3))

    assert ds(Kd(i, k), (k, -oo, oo)) == 1
    assert ds(Kd(i, k), (k, 0, oo)) == Piecewise((1, Integer(0) <= i), (0, True))
    assert ds(Kd(i, k), (k, 1, 3)) == \
        Piecewise((1, And(Integer(1) <= i, i <= 3)), (0, True))
    assert ds(k*Kd(i, j)*Kd(j, k), (k, -oo, oo)) == j*Kd(i, j)
    assert ds(j*Kd(i, j), (j, -oo, oo)) == i
    assert ds(i*Kd(i, j), (i, -oo, oo)) == j
    assert ds(x, (i, 1, 3)) == 3*x
    assert ds((i + j)*Kd(i, j), (j, -oo, oo)) == 2*i


def test_deltasummation_basic_symbolic():
    assert ds(Kd(exp(i), 0), (i, 1, 3)) == 0
    assert ds(Kd(exp(i), 0), (i, -1, 3)) == 0
    assert ds(Kd(exp(i), 1), (i, 0, 3)) == 1
    assert ds(Kd(exp(i), 1), (i, 1, 3)) == 0
    assert ds(Kd(exp(i), 1), (i, -10, 3)) == 1
    assert ds(Kd(i, j), (j, 1, 3)) == \
        Piecewise((1, And(Integer(1) <= i, i <= 3)), (0, True))
    assert ds(Kd(i, j), (j, 1, 1)) == Piecewise((1, Eq(i, 1)), (0, True))
    assert ds(Kd(i, j), (j, 2, 2)) == Piecewise((1, Eq(i, 2)), (0, True))
    assert ds(Kd(i, j), (j, 3, 3)) == Piecewise((1, Eq(i, 3)), (0, True))
    assert ds(Kd(i, j), (j, 1, k)) == \
        Piecewise((1, And(Integer(1) <= i, i <= k)), (0, True))
    assert ds(Kd(i, j), (j, k, 3)) == \
        Piecewise((1, And(k <= i, i <= 3)), (0, True))
    assert ds(Kd(i, j), (j, k, l)) == \
        Piecewise((1, And(k <= i, i <= l)), (0, True))


def test_deltasummation_mul_x_kd():
    assert ds(x*Kd(i, j), (j, 1, 3)) == \
        Piecewise((x, And(Integer(1) <= i, i <= 3)), (0, True))
    assert ds(x*Kd(i, j), (j, 1, 1)) == Piecewise((x, Eq(i, 1)), (0, True))
    assert ds(x*Kd(i, j), (j, 2, 2)) == Piecewise((x, Eq(i, 2)), (0, True))
    assert ds(x*Kd(i, j), (j, 3, 3)) == Piecewise((x, Eq(i, 3)), (0, True))
    assert ds(x*Kd(i, j), (j, 1, k)) == \
        Piecewise((x, And(Integer(1) <= i, i <= k)), (0, True))
    assert ds(x*Kd(i, j), (j, k, 3)) == \
        Piecewise((x, And(k <= i, i <= 3)), (0, True))
    assert ds(x*Kd(i, j), (j, k, l)) == \
        Piecewise((x, And(k <= i, i <= l)), (0, True))


def test_deltasummation_mul_add_x_y_kd():
    assert ds((x + y)*Kd(i, j), (j, 1, 3)) == \
        Piecewise((x + y, And(Integer(1) <= i, i <= 3)), (0, True))
    assert ds((x + y)*Kd(i, j), (j, 1, 1)) == \
        Piecewise((x + y, Eq(i, 1)), (0, True))
    assert ds((x + y)*Kd(i, j), (j, 2, 2)) == \
        Piecewise((x + y, Eq(i, 2)), (0, True))
    assert ds((x + y)*Kd(i, j), (j, 3, 3)) == \
        Piecewise((x + y, Eq(i, 3)), (0, True))
    assert ds((x + y)*Kd(i, j), (j, 1, k)) == \
        Piecewise((x + y, And(Integer(1) <= i, i <= k)), (0, True))
    assert ds((x + y)*Kd(i, j), (j, k, 3)) == \
        Piecewise((x + y, And(k <= i, i <= 3)), (0, True))
    assert ds((x + y)*Kd(i, j), (j, k, l)) == \
        Piecewise((x + y, And(k <= i, i <= l)), (0, True))


def test_deltasummation_add_kd_kd():
    assert ds(Kd(i, k) + Kd(j, k), (k, 1, 3)) == piecewise_fold(
        Piecewise((1, And(Integer(1) <= i, i <= 3)), (0, True)) +
        Piecewise((1, And(Integer(1) <= j, j <= 3)), (0, True)))
    assert ds(Kd(i, k) + Kd(j, k), (k, 1, 1)) == piecewise_fold(
        Piecewise((1, Eq(i, 1)), (0, True)) +
        Piecewise((1, Eq(j, 1)), (0, True)))
    assert ds(Kd(i, k) + Kd(j, k), (k, 2, 2)) == piecewise_fold(
        Piecewise((1, Eq(i, 2)), (0, True)) +
        Piecewise((1, Eq(j, 2)), (0, True)))
    assert ds(Kd(i, k) + Kd(j, k), (k, 3, 3)) == piecewise_fold(
        Piecewise((1, Eq(i, 3)), (0, True)) +
        Piecewise((1, Eq(j, 3)), (0, True)))
    assert ds(Kd(i, k) + Kd(j, k), (k, 1, l)) == piecewise_fold(
        Piecewise((1, And(Integer(1) <= i, i <= l)), (0, True)) +
        Piecewise((1, And(Integer(1) <= j, j <= l)), (0, True)))
    assert ds(Kd(i, k) + Kd(j, k), (k, l, 3)) == piecewise_fold(
        Piecewise((1, And(l <= i, i <= 3)), (0, True)) +
        Piecewise((1, And(l <= j, j <= 3)), (0, True)))
    assert ds(Kd(i, k) + Kd(j, k), (k, l, m)) == piecewise_fold(
        Piecewise((1, And(l <= i, i <= m)), (0, True)) +
        Piecewise((1, And(l <= j, j <= m)), (0, True)))


def test_deltasummation_add_mul_x_kd_kd():
    assert ds(x*Kd(i, k) + Kd(j, k), (k, 1, 3)) == piecewise_fold(
        Piecewise((x, And(Integer(1) <= i, i <= 3)), (0, True)) +
        Piecewise((1, And(Integer(1) <= j, j <= 3)), (0, True)))
    assert ds(x*Kd(i, k) + Kd(j, k), (k, 1, 1)) == piecewise_fold(
        Piecewise((x, Eq(i, 1)), (0, True)) +
        Piecewise((1, Eq(j, 1)), (0, True)))
    assert ds(x*Kd(i, k) + Kd(j, k), (k, 2, 2)) == piecewise_fold(
        Piecewise((x, Eq(i, 2)), (0, True)) +
        Piecewise((1, Eq(j, 2)), (0, True)))
    assert ds(x*Kd(i, k) + Kd(j, k), (k, 3, 3)) == piecewise_fold(
        Piecewise((x, Eq(i, 3)), (0, True)) +
        Piecewise((1, Eq(j, 3)), (0, True)))
    assert ds(x*Kd(i, k) + Kd(j, k), (k, 1, l)) == piecewise_fold(
        Piecewise((x, And(Integer(1) <= i, i <= l)), (0, True)) +
        Piecewise((1, And(Integer(1) <= j, j <= l)), (0, True)))
    assert ds(x*Kd(i, k) + Kd(j, k), (k, l, 3)) == piecewise_fold(
        Piecewise((x, And(l <= i, i <= 3)), (0, True)) +
        Piecewise((1, And(l <= j, j <= 3)), (0, True)))
    assert ds(x*Kd(i, k) + Kd(j, k), (k, l, m)) == piecewise_fold(
        Piecewise((x, And(l <= i, i <= m)), (0, True)) +
        Piecewise((1, And(l <= j, j <= m)), (0, True)))


def test_deltasummation_mul_x_add_kd_kd():
    assert ds(x*(Kd(i, k) + Kd(j, k)), (k, 1, 3)) == piecewise_fold(
        Piecewise((x, And(Integer(1) <= i, i <= 3)), (0, True)) +
        Piecewise((x, And(Integer(1) <= j, j <= 3)), (0, True)))
    assert ds(x*(Kd(i, k) + Kd(j, k)), (k, 1, 1)) == piecewise_fold(
        Piecewise((x, Eq(i, 1)), (0, True)) +
        Piecewise((x, Eq(j, 1)), (0, True)))
    assert ds(x*(Kd(i, k) + Kd(j, k)), (k, 2, 2)) == piecewise_fold(
        Piecewise((x, Eq(i, 2)), (0, True)) +
        Piecewise((x, Eq(j, 2)), (0, True)))
    assert ds(x*(Kd(i, k) + Kd(j, k)), (k, 3, 3)) == piecewise_fold(
        Piecewise((x, Eq(i, 3)), (0, True)) +
        Piecewise((x, Eq(j, 3)), (0, True)))
    assert ds(x*(Kd(i, k) + Kd(j, k)), (k, 1, l)) == piecewise_fold(
        Piecewise((x, And(Integer(1) <= i, i <= l)), (0, True)) +
        Piecewise((x, And(Integer(1) <= j, j <= l)), (0, True)))
    assert ds(x*(Kd(i, k) + Kd(j, k)), (k, l, 3)) == piecewise_fold(
        Piecewise((x, And(l <= i, i <= 3)), (0, True)) +
        Piecewise((x, And(l <= j, j <= 3)), (0, True)))
    assert ds(x*(Kd(i, k) + Kd(j, k)), (k, l, m)) == piecewise_fold(
        Piecewise((x, And(l <= i, i <= m)), (0, True)) +
        Piecewise((x, And(l <= j, j <= m)), (0, True)))


def test_deltasummation_mul_add_x_y_add_kd_kd():
    assert ds((x + y)*(Kd(i, k) + Kd(j, k)), (k, 1, 3)) == piecewise_fold(
        Piecewise((x + y, And(Integer(1) <= i, i <= 3)), (0, True)) +
        Piecewise((x + y, And(Integer(1) <= j, j <= 3)), (0, True)))
    assert ds((x + y)*(Kd(i, k) + Kd(j, k)), (k, 1, 1)) == piecewise_fold(
        Piecewise((x + y, Eq(i, 1)), (0, True)) +
        Piecewise((x + y, Eq(j, 1)), (0, True)))
    assert ds((x + y)*(Kd(i, k) + Kd(j, k)), (k, 2, 2)) == piecewise_fold(
        Piecewise((x + y, Eq(i, 2)), (0, True)) +
        Piecewise((x + y, Eq(j, 2)), (0, True)))
    assert ds((x + y)*(Kd(i, k) + Kd(j, k)), (k, 3, 3)) == piecewise_fold(
        Piecewise((x + y, Eq(i, 3)), (0, True)) +
        Piecewise((x + y, Eq(j, 3)), (0, True)))
    assert ds((x + y)*(Kd(i, k) + Kd(j, k)), (k, 1, l)) == piecewise_fold(
        Piecewise((x + y, And(Integer(1) <= i, i <= l)), (0, True)) +
        Piecewise((x + y, And(Integer(1) <= j, j <= l)), (0, True)))
    assert ds((x + y)*(Kd(i, k) + Kd(j, k)), (k, l, 3)) == piecewise_fold(
        Piecewise((x + y, And(l <= i, i <= 3)), (0, True)) +
        Piecewise((x + y, And(l <= j, j <= 3)), (0, True)))
    assert ds((x + y)*(Kd(i, k) + Kd(j, k)), (k, l, m)) == piecewise_fold(
        Piecewise((x + y, And(l <= i, i <= m)), (0, True)) +
        Piecewise((x + y, And(l <= j, j <= m)), (0, True)))


def test_deltasummation_add_mul_x_y_mul_x_kd():
    assert ds(x*y + x*Kd(i, j), (j, 1, 3)) == \
        Piecewise((3*x*y + x, And(Integer(1) <= i, i <= 3)), (3*x*y, True))
    assert ds(x*y + x*Kd(i, j), (j, 1, 1)) == \
        Piecewise((x*y + x, Eq(i, 1)), (x*y, True))
    assert ds(x*y + x*Kd(i, j), (j, 2, 2)) == \
        Piecewise((x*y + x, Eq(i, 2)), (x*y, True))
    assert ds(x*y + x*Kd(i, j), (j, 3, 3)) == \
        Piecewise((x*y + x, Eq(i, 3)), (x*y, True))
    assert ds(x*y + x*Kd(i, j), (j, 1, k)) == \
        Piecewise((k*x*y + x, And(Integer(1) <= i, i <= k)), (k*x*y, True))
    assert ds(x*y + x*Kd(i, j), (j, k, 3)) == \
        Piecewise(((4 - k)*x*y + x, And(k <= i, i <= 3)), ((4 - k)*x*y, True))
    assert ds(x*y + x*Kd(i, j), (j, k, l)) == Piecewise(
        ((l - k + 1)*x*y + x, And(k <= i, i <= l)), ((l - k + 1)*x*y, True))


def test_deltasummation_mul_x_add_y_kd():
    assert ds(x*(y + Kd(i, j)), (j, 1, 3)) == \
        Piecewise((3*x*y + x, And(Integer(1) <= i, i <= 3)), (3*x*y, True))
    assert ds(x*(y + Kd(i, j)), (j, 1, 1)) == \
        Piecewise((x*y + x, Eq(i, 1)), (x*y, True))
    assert ds(x*(y + Kd(i, j)), (j, 2, 2)) == \
        Piecewise((x*y + x, Eq(i, 2)), (x*y, True))
    assert ds(x*(y + Kd(i, j)), (j, 3, 3)) == \
        Piecewise((x*y + x, Eq(i, 3)), (x*y, True))
    assert ds(x*(y + Kd(i, j)), (j, 1, k)) == \
        Piecewise((k*x*y + x, And(Integer(1) <= i, i <= k)), (k*x*y, True))
    assert ds(x*(y + Kd(i, j)), (j, k, 3)) == \
        Piecewise(((4 - k)*x*y + x, And(k <= i, i <= 3)), ((4 - k)*x*y, True))
    assert ds(x*(y + Kd(i, j)), (j, k, l)) == Piecewise(
        ((l - k + 1)*x*y + x, And(k <= i, i <= l)), ((l - k + 1)*x*y, True))


def test_deltasummation_mul_x_add_y_twokd():
    assert ds(x*(y + 2*Kd(i, j)), (j, 1, 3)) == \
        Piecewise((3*x*y + 2*x, And(Integer(1) <= i, i <= 3)), (3*x*y, True))
    assert ds(x*(y + 2*Kd(i, j)), (j, 1, 1)) == \
        Piecewise((x*y + 2*x, Eq(i, 1)), (x*y, True))
    assert ds(x*(y + 2*Kd(i, j)), (j, 2, 2)) == \
        Piecewise((x*y + 2*x, Eq(i, 2)), (x*y, True))
    assert ds(x*(y + 2*Kd(i, j)), (j, 3, 3)) == \
        Piecewise((x*y + 2*x, Eq(i, 3)), (x*y, True))
    assert ds(x*(y + 2*Kd(i, j)), (j, 1, k)) == \
        Piecewise((k*x*y + 2*x, And(Integer(1) <= i, i <= k)), (k*x*y, True))
    assert ds(x*(y + 2*Kd(i, j)), (j, k, 3)) == Piecewise(
        ((4 - k)*x*y + 2*x, And(k <= i, i <= 3)), ((4 - k)*x*y, True))
    assert ds(x*(y + 2*Kd(i, j)), (j, k, l)) == Piecewise(
        ((l - k + 1)*x*y + 2*x, And(k <= i, i <= l)), ((l - k + 1)*x*y, True))


def test_deltasummation_mul_add_x_y_add_y_kd():
    assert ds((x + y)*(y + Kd(i, j)), (j, 1, 3)) == Piecewise(
        (3*(x + y)*y + x + y, And(Integer(1) <= i, i <= 3)), (3*(x + y)*y, True))
    assert ds((x + y)*(y + Kd(i, j)), (j, 1, 1)) == \
        Piecewise(((x + y)*y + x + y, Eq(i, 1)), ((x + y)*y, True))
    assert ds((x + y)*(y + Kd(i, j)), (j, 2, 2)) == \
        Piecewise(((x + y)*y + x + y, Eq(i, 2)), ((x + y)*y, True))
    assert ds((x + y)*(y + Kd(i, j)), (j, 3, 3)) == \
        Piecewise(((x + y)*y + x + y, Eq(i, 3)), ((x + y)*y, True))
    assert ds((x + y)*(y + Kd(i, j)), (j, 1, k)) == Piecewise(
        (k*(x + y)*y + x + y, And(Integer(1) <= i, i <= k)), (k*(x + y)*y, True))
    assert ds((x + y)*(y + Kd(i, j)), (j, k, 3)) == Piecewise(
        ((4 - k)*(x + y)*y + x + y, And(k <= i, i <= 3)),
        ((4 - k)*(x + y)*y, True))
    assert ds((x + y)*(y + Kd(i, j)), (j, k, l)) == Piecewise(
        ((l - k + 1)*(x + y)*y + x + y, And(k <= i, i <= l)),
        ((l - k + 1)*(x + y)*y, True))


def test_deltasummation_mul_add_x_kd_add_y_kd():
    assert ds((x + Kd(i, k))*(y + Kd(i, j)), (j, 1, 3)) == piecewise_fold(
        Piecewise((Kd(i, k) + x, And(Integer(1) <= i, i <= 3)), (0, True)) +
        3*(Kd(i, k) + x)*y)
    assert ds((x + Kd(i, k))*(y + Kd(i, j)), (j, 1, 1)) == piecewise_fold(
        Piecewise((Kd(i, k) + x, Eq(i, 1)), (0, True)) +
        (Kd(i, k) + x)*y)
    assert ds((x + Kd(i, k))*(y + Kd(i, j)), (j, 2, 2)) == piecewise_fold(
        Piecewise((Kd(i, k) + x, Eq(i, 2)), (0, True)) +
        (Kd(i, k) + x)*y)
    assert ds((x + Kd(i, k))*(y + Kd(i, j)), (j, 3, 3)) == piecewise_fold(
        Piecewise((Kd(i, k) + x, Eq(i, 3)), (0, True)) +
        (Kd(i, k) + x)*y)
    assert ds((x + Kd(i, k))*(y + Kd(i, j)), (j, 1, k)) == piecewise_fold(
        Piecewise((Kd(i, k) + x, And(Integer(1) <= i, i <= k)), (0, True)) +
        k*(Kd(i, k) + x)*y)
    assert ds((x + Kd(i, k))*(y + Kd(i, j)), (j, k, 3)) == piecewise_fold(
        Piecewise((Kd(i, k) + x, And(k <= i, i <= 3)), (0, True)) +
        (4 - k)*(Kd(i, k) + x)*y)
    assert ds((x + Kd(i, k))*(y + Kd(i, j)), (j, k, l)) == piecewise_fold(
        Piecewise((Kd(i, k) + x, And(k <= i, i <= l)), (0, True)) +
        (l - k + 1)*(Kd(i, k) + x)*y)



"""Geometrical Points.

Contains
========
Point
Point2D
Point3D
"""

from ..core import Add, Float, S, Tuple, sympify
from ..core.compatibility import iterable
from ..core.evaluate import global_evaluate
from ..functions import im, sqrt
from ..matrices import Matrix
from ..simplify import nsimplify, simplify
from .entity import GeometryEntity
from .exceptions import GeometryError


class Point(GeometryEntity):
    """A point in a n-dimensional Euclidean space.

    Parameters
    ==========

    coords : sequence of n-coordinate values. In the special
    case where n=2 or 3, a Point2D or Point3D will be created
    as appropriate.

    Attributes
    ==========

    length
    origin: A `Point` representing the origin of the
        appropriately-dimensioned space.

    Raises
    ======

    TypeError
        When trying to add or subtract points with different dimensions.
        When `intersection` is called with object other than a Point.

    See Also
    ========

    diofant.geometry.line.Segment : Connects two Points

    Examples
    ========

    >>> from diofant.geometry import Point
    >>> from diofant.abc import x
    >>> Point(1, 2, 3)
    Point3D(1, 2, 3)
    >>> Point([1, 2])
    Point2D(1, 2)
    >>> Point(0, x)
    Point2D(0, x)

    Floats are automatically converted to Rational unless the
    evaluate flag is False:

    >>> Point(0.5, 0.25)
    Point2D(1/2, 1/4)
    >>> print(Point(0.5, 0.25, evaluate=False))
    Point2D(0.5, 0.25)
    """

    def __new__(cls, *args, **kwargs):
        evaluate = kwargs.get('evaluate', global_evaluate[0])

        if iterable(args[0]):
            if isinstance(args[0], Point) and not evaluate:
                return args[0]
            args = args[0]

        # unpack the arguments into a friendly Tuple
        # if we were already a Point, we're doing an excess
        # iteration, but we'll worry about efficiency later
        coords = Tuple(*args)
        if any(a.is_number and im(a) for a in coords):
            raise ValueError('Imaginary coordinates not permitted.')

        # Turn any Floats into rationals and simplify
        # any expressions before we instantiate
        if evaluate:
            coords = coords.xreplace({f: simplify(nsimplify(f, rational=True))
                                      for f in coords.atoms(Float)})
        if len(coords) == 2:
            return Point2D(coords, **kwargs)
        if len(coords) == 3:
            return Point3D(coords, **kwargs)

        return GeometryEntity.__new__(cls, *coords)

    def __contains__(self, item):
        return item == self

    def is_concyclic(*args):
        # Coincident points are irrelevant and can confuse this algorithm.
        # Use only unique points.
        args = list(set(args))
        if not all(isinstance(p, Point) for p in args):
            raise TypeError('Must pass only Point objects')

        return args[0].is_concyclic(*args[1:])

    def is_collinear(*args):
        """Is a sequence of points collinear?

        Test whether or not a set of points are collinear. Returns True if
        the set of points are collinear, or False otherwise.

        Parameters
        ==========

        points : sequence of Point

        Returns
        =======

        is_collinear : boolean

        Notes
        =====

        Slope is preserved everywhere on a line, so the slope between
        any two points on the line should be the same. Take the first
        two points, p1 and p2, and create a translated point v1
        with p1 as the origin. Now for every other point we create
        a translated point, vi with p1 also as the origin. Note that
        these translations preserve slope since everything is
        consistently translated to a new origin of p1. Since slope
        is preserved then we have the following equality:

              * v1_slope = vi_slope
              * v1.y/v1.x = vi.y/vi.x (due to translation)
              * v1.y*vi.x = vi.y*v1.x
              * v1.y*vi.x - vi.y*v1.x = 0           (*)

        Hence, if we have a vi such that the equality in (*) is False
        then the points are not collinear. We do this test for every
        point in the list, and if all pass then they are collinear.

        See Also
        ========

        diofant.geometry.line.Line

        Examples
        ========

        >>> from diofant import Point
        >>> from diofant.abc import x
        >>> p1, p2 = Point(0, 0), Point(1, 1)
        >>> p3, p4, p5 = Point(2, 2), Point(x, x), Point(1, 2)
        >>> Point.is_collinear(p1, p2, p3, p4)
        True
        >>> Point.is_collinear(p1, p2, p3, p5)
        False

        """

        # Coincident points are irrelevant; use only unique points.
        args = list(set(args))
        if not all(isinstance(p, Point) for p in args):
            raise TypeError('Must pass only Point objects')

        if len(args) == 0:
            return False
        if len(args) <= 2:
            return True

        # translate our points
        points = [p - args[0] for p in args[1:]]
        for p in points[1:]:
            if not Point.is_scalar_multiple(points[0], p):
                return False
        return True

    def is_scalar_multiple(self, other):
        """Returns whether `self` and `other` are scalar multiples
        of each other.
        """
        # if the vectors self and other are linearly dependent, then they must
        # be scalar multiples of each other
        m = Matrix([self.args, other.args])
        return m.rank() < 2

    @property
    def length(self):
        """
        Treating a Point as a Line, this returns 0 for the length of a Point.

        Examples
        ========

        >>> from diofant import Point
        >>> p = Point(0, 1)
        >>> p.length
        0
        """
        return S.Zero

    @property
    def origin(self):
        """A point of all zeros of the same ambient dimension
        as the current point
        """
        return Point([0]*len(self))

    @property
    def is_zero(self):
        """True if every coordinate is zero, otherwise False."""
        return all(x == 0 for x in self.args)

    @property
    def ambient_dimension(self):
        """The dimension of the ambient space the point is in.
        I.e., if the point is in R^n, the ambient dimension
        will be n
        """
        return len(self)

    def distance(self, p):
        """The Euclidean distance from self to point p.

        Parameters
        ==========

        p : Point

        Returns
        =======

        distance : number or symbolic expression.

        See Also
        ========

        diofant.geometry.line.Segment.length

        Examples
        ========

        >>> from diofant.geometry import Point
        >>> p1, p2 = Point(1, 1), Point(4, 5)
        >>> p1.distance(p2)
        5

        >>> from diofant.abc import x, y
        >>> p3 = Point(x, y)
        >>> p3.distance(Point(0, 0))
        sqrt(x**2 + y**2)

        """
        p = Point(p)
        return sqrt(sum((a - b)**2
                        for a, b in zip(self.args,
                                        p.args if isinstance(p, Point) else p)))

    def midpoint(self, p):
        """The midpoint between self and point p.

        Parameters
        ==========

        p : Point

        Returns
        =======

        midpoint : Point

        See Also
        ========

        diofant.geometry.line.Segment.midpoint

        Examples
        ========

        >>> from diofant.geometry import Point
        >>> p1, p2 = Point(1, 1), Point(13, 5)
        >>> p1.midpoint(p2)
        Point2D(7, 3)

        """
        return Point([simplify((a + b)*S.Half) for a, b in zip(self.args, p.args)])

    def evalf(self, prec=None, **options):
        """Evaluate the coordinates of the point.

        This method will, where possible, create and return a new Point
        where the coordinates are evaluated as floating point numbers to
        the precision indicated (default=15).

        Returns
        =======

        point : Point

        Examples
        ========

        >>> from diofant import Point, Rational
        >>> p1 = Point(Rational(1, 2), Rational(3, 2))
        >>> p1
        Point2D(1/2, 3/2)
        >>> print(p1.evalf())
        Point2D(0.5, 1.5)
        """
        coords = [x.evalf(prec, **options) for x in self.args]
        return Point(*coords, evaluate=False)

    n = evalf

    def intersection(self, o):
        """The intersection between this point and another point.

        Parameters
        ==========

        other : Point

        Returns
        =======

        intersection : list of Points

        Notes
        =====

        The return value will either be an empty list if there is no
        intersection, otherwise it will contain this point.

        Examples
        ========

        >>> from diofant import Point
        >>> p1, p2, p3 = Point(0, 0), Point(1, 1), Point(0, 0)
        >>> p1.intersection(p2)
        []
        >>> p1.intersection(p3)
        [Point2D(0, 0)]

        """
        if isinstance(o, Point):
            if len(self) != len(o):
                raise ValueError("Points must be of the same dimension to intersect")
            if self == o:
                return [self]
            return []

        return o.intersection(self)

    def dot(self, p2):
        """Return dot product of self with another Point."""
        p2 = Point(p2)
        return Add(*[a*b for a, b in zip(self, p2)])

    def equals(self, other):
        """Returns whether the coordinates of self and other agree."""
        # a point is equal to another point if all its components are equal
        if not isinstance(other, Point) or len(self.args) != len(other.args):
            return False
        return all(a.equals(b) for a, b in zip(self.args, other.args))

    def __len__(self):
        return len(self.args)

    def __iter__(self):
        return self.args.__iter__()

    def __getitem__(self, key):
        return self.args[key]

    def __add__(self, other):
        """Add other to self by incrementing self's coordinates by those of other.

        See Also
        ========

        diofant.geometry.entity.GeometryEntity.translate
        """

        if iterable(other) and len(other) == len(self):
            return Point([simplify(a + b) for a, b in zip(self, other)])
        else:
            raise ValueError(
                "Points must have the same number of dimensions")

    def __sub__(self, other):
        """Subtract two points, or subtract a factor from this point's
        coordinates.
        """
        return self + (-other)

    def __mul__(self, factor):
        """Multiply point's coordinates by a factor."""
        factor = sympify(factor)
        return Point([simplify(x*factor) for x in self.args])

    def __truediv__(self, divisor):
        """Divide point's coordinates by a factor."""
        divisor = sympify(divisor)
        return Point([simplify(x/divisor) for x in self.args])

    def __neg__(self):
        """Negate the point."""
        return Point([-x for x in self.args])

    def __abs__(self):
        """Returns the distance between this point and the origin."""
        origin = Point([0]*len(self))
        return Point.distance(origin, self)


class Point2D(Point):
    """A point in a 2-dimensional Euclidean space.

    Parameters
    ==========

    coords : sequence of 2 coordinate values.

    Attributes
    ==========

    x
    y
    diofant.geometry.Point.length

    Raises
    ======

    TypeError
        When trying to add or subtract points with different dimensions.
        When trying to create a point with more than two dimensions.
        When `intersection` is called with object other than a Point.

    See Also
    ========

    diofant.geometry.line.Segment : Connects two Points

    Examples
    ========

    >>> from diofant.geometry import Point2D
    >>> from diofant.abc import x
    >>> Point2D(1, 2)
    Point2D(1, 2)
    >>> Point2D([1, 2])
    Point2D(1, 2)
    >>> Point2D(0, x)
    Point2D(0, x)

    Floats are automatically converted to Rational unless the
    evaluate flag is False:

    >>> Point2D(0.5, 0.25)
    Point2D(1/2, 1/4)
    >>> print(Point2D(0.5, 0.25, evaluate=False))
    Point2D(0.5, 0.25)
    """

    def __new__(cls, *args, **kwargs):
        eval = kwargs.get('evaluate', global_evaluate[0])
        check = True
        if isinstance(args[0], Point2D):
            if not eval:
                return args[0]
            args = args[0].args
            check = False
        else:
            if iterable(args[0]):
                args = args[0]
            if len(args) != 2:
                raise ValueError(
                    "Only two dimensional points currently supported")
        coords = Tuple(*args)
        if check:
            if any(a.is_number and im(a) for a in coords):
                raise ValueError('Imaginary args not permitted.')
        if eval:
            coords = coords.xreplace({f: simplify(nsimplify(f, rational=True))
                                      for f in coords.atoms(Float)})
        return GeometryEntity.__new__(cls, *coords)

    def __contains__(self, item):
        return item == self

    @property
    def x(self):
        """
        Returns the X coordinate of the Point.

        Examples
        ========

        >>> from diofant import Point2D
        >>> p = Point2D(0, 1)
        >>> p.x
        0
        """
        return self.args[0]

    @property
    def y(self):
        """
        Returns the Y coordinate of the Point.

        Examples
        ========

        >>> from diofant import Point2D
        >>> p = Point2D(0, 1)
        >>> p.y
        1
        """
        return self.args[1]

    @property
    def bounds(self):
        """Return a tuple (xmin, ymin, xmax, ymax) representing the bounding
        rectangle for the geometric figure.

        """

        return self.x, self.y, self.x, self.y

    def is_concyclic(*points):
        """Is a sequence of points concyclic?

        Test whether or not a sequence of points are concyclic (i.e., they lie
        on a circle).

        Parameters
        ==========

        points : sequence of Points

        Returns
        =======

        is_concyclic : boolean
            True if points are concyclic, False otherwise.

        See Also
        ========

        diofant.geometry.ellipse.Circle

        Notes
        =====

        No points are not considered to be concyclic. One or two points
        are definitely concyclic and three points are conyclic iff they
        are not collinear.

        For more than three points, create a circle from the first three
        points. If the circle cannot be created (i.e., they are collinear)
        then all of the points cannot be concyclic. If the circle is created
        successfully then simply check the remaining points for containment
        in the circle.

        Examples
        ========

        >>> from diofant.geometry import Point
        >>> p1, p2 = Point(-1, 0), Point(1, 0)
        >>> p3, p4 = Point(0, 1), Point(-1, 2)
        >>> Point.is_concyclic(p1, p2, p3)
        True
        >>> Point.is_concyclic(p1, p2, p3, p4)
        False

        """
        if len(points) == 0:
            return False
        if len(points) <= 2:
            return True
        points = [Point(p) for p in points]
        if len(points) == 3:
            return (not Point.is_collinear(*points))

        try:
            from .ellipse import Circle
            c = Circle(points[0], points[1], points[2])
            for point in points[3:]:
                if point not in c:
                    return False
            return True
        except GeometryError:
            # Circle could not be created, because of collinearity of the
            # three points passed in, hence they are not concyclic.
            return False

    def rotate(self, angle, pt=None):
        """Rotate ``angle`` radians counterclockwise about Point ``pt``.

        See Also
        ========

        rotate, scale

        Examples
        ========

        >>> from diofant import Point2D, pi
        >>> t = Point2D(1, 0)
        >>> t.rotate(pi/2)
        Point2D(0, 1)
        >>> t.rotate(pi/2, (2, 0))
        Point2D(2, -1)
        """
        from ..functions import cos, sin

        c = cos(angle)
        s = sin(angle)

        rv = self
        if pt is not None:
            pt = Point(pt)
            rv -= pt
        x, y = rv.args
        rv = Point(c*x - s*y, s*x + c*y)
        if pt is not None:
            rv += pt
        return rv

    def scale(self, x=1, y=1, pt=None):
        """Scale the coordinates of the Point by multiplying by
        ``x`` and ``y`` after subtracting ``pt`` -- default is (0, 0) --
        and then adding ``pt`` back again (i.e. ``pt`` is the point of
        reference for the scaling).

        See Also
        ========

        rotate, translate

        Examples
        ========

        >>> from diofant import Point2D
        >>> t = Point2D(1, 1)
        >>> t.scale(2)
        Point2D(2, 1)
        >>> t.scale(2, 2)
        Point2D(2, 2)

        """
        if pt:
            pt = Point(pt)
            return self.translate(*(-pt).args).scale(x, y).translate(*pt.args)
        return Point(self.x*x, self.y*y)

    def translate(self, x=0, y=0):
        """Shift the Point by adding x and y to the coordinates of the Point.

        See Also
        ========

        rotate, scale

        Examples
        ========

        >>> from diofant import Point2D
        >>> t = Point2D(0, 1)
        >>> t.translate(2)
        Point2D(2, 1)
        >>> t.translate(2, 2)
        Point2D(2, 3)
        >>> t + Point2D(2, 2)
        Point2D(2, 3)

        """
        return Point(self.x + x, self.y + y)

    def transform(self, matrix):
        """Return the point after applying the transformation described
        by the 3x3 Matrix, ``matrix``.

        See Also
        ========

        diofant.geometry.entity.GeometryEntity.rotate
        diofant.geometry.entity.GeometryEntity.scale
        diofant.geometry.entity.GeometryEntity.translate
        """
        try:
            col, row = matrix.shape
            valid_matrix = matrix.is_square and col == 3
        except AttributeError:
            # We hit this block if matrix argument is not actually a Matrix.
            valid_matrix = False
        if not valid_matrix:
            raise ValueError("The argument to the transform function must be "
                             + "a 3x3 matrix")
        x, y = self.args
        return Point(*(Matrix(1, 3, [x, y, 1])*matrix).tolist()[0][:2])


class Point3D(Point):
    """A point in a 3-dimensional Euclidean space.

    Parameters
    ==========

    coords : sequence of 3 coordinate values.

    Attributes
    ==========

    x
    y
    z
    diofant.geometry.Point.length

    Raises
    ======

    TypeError
        When trying to add or subtract points with different dimensions.
        When `intersection` is called with object other than a Point.

    Notes
    =====

    Currently only 2-dimensional and 3-dimensional points are supported.

    Examples
    ========

    >>> from diofant import Point3D
    >>> from diofant.abc import x
    >>> Point3D(1, 2, 3)
    Point3D(1, 2, 3)
    >>> Point3D([1, 2, 3])
    Point3D(1, 2, 3)
    >>> Point3D(0, x, 3)
    Point3D(0, x, 3)

    Floats are automatically converted to Rational unless the
    evaluate flag is False:

    >>> Point3D(0.5, 0.25, 2)
    Point3D(1/2, 1/4, 2)
    >>> print(Point3D(0.5, 0.25, 3, evaluate=False))
    Point3D(0.5, 0.25, 3)
    """

    def __new__(cls, *args, **kwargs):
        eval = kwargs.get('evaluate', global_evaluate[0])
        if isinstance(args[0], (Point, Point3D)):
            if not eval:
                return args[0]
            args = args[0].args
        else:
            if iterable(args[0]):
                args = args[0]
            if len(args) not in (2, 3):
                raise TypeError(
                    "Enter a 2 or 3 dimensional point")
        coords = Tuple(*args)
        if len(coords) == 2:
            coords += (S.Zero,)
        if eval:
            coords = coords.xreplace({f: simplify(nsimplify(f, rational=True))
                                      for f in coords.atoms(Float)})
        return GeometryEntity.__new__(cls, *coords)

    def __contains__(self, item):
        return item == self

    @property
    def x(self):
        """
        Returns the X coordinate of the Point.

        Examples
        ========

        >>> from diofant import Point3D
        >>> p = Point3D(0, 1, 3)
        >>> p.x
        0
        """
        return self.args[0]

    @property
    def y(self):
        """
        Returns the Y coordinate of the Point.

        Examples
        ========

        >>> from diofant import Point3D
        >>> p = Point3D(0, 1, 2)
        >>> p.y
        1
        """
        return self.args[1]

    @property
    def z(self):
        """
        Returns the Z coordinate of the Point.

        Examples
        ========

        >>> from diofant import Point3D
        >>> p = Point3D(0, 1, 1)
        >>> p.z
        1
        """
        return self.args[2]

    def direction_ratio(self, point):
        """
        Gives the direction ratio between 2 points

        Parameters
        ==========

        p : Point3D

        Returns
        =======

        list

        Examples
        ========

        >>> from diofant import Point3D
        >>> p1 = Point3D(1, 2, 3)
        >>> p1.direction_ratio(Point3D(2, 3, 5))
        [1, 1, 2]
        """
        return [(point.x - self.x), (point.y - self.y), (point.z - self.z)]

    def direction_cosine(self, point):
        """
        Gives the direction cosine between 2 points

        Parameters
        ==========

        p : Point3D

        Returns
        =======

        list

        Examples
        ========

        >>> from diofant import Point3D
        >>> p1 = Point3D(1, 2, 3)
        >>> p1.direction_cosine(Point3D(2, 3, 5))
        [sqrt(6)/6, sqrt(6)/6, sqrt(6)/3]
        """
        a = self.direction_ratio(point)
        b = sqrt(sum(i**2 for i in a))
        return [(point.x - self.x) / b, (point.y - self.y) / b,
                (point.z - self.z) / b]

    @staticmethod
    def are_collinear(*points):
        """Is a sequence of points collinear?

        Test whether or not a set of points are collinear. Returns True if
        the set of points are collinear, or False otherwise.

        Parameters
        ==========

        points : sequence of Point

        Returns
        =======

        are_collinear : boolean

        See Also
        ========

        diofant.geometry.line3d.Line3D

        Examples
        ========

        >>> from diofant import Point3D, Matrix
        >>> from diofant.abc import x
        >>> p1, p2 = Point3D(0, 0, 0), Point3D(1, 1, 1)
        >>> p3, p4, p5 = Point3D(2, 2, 2), Point3D(x, x, x), Point3D(1, 2, 6)
        >>> Point3D.are_collinear(p1, p2, p3, p4)
        True
        >>> Point3D.are_collinear(p1, p2, p3, p5)
        False
        """
        return Point.is_collinear(*points)

    @staticmethod
    def are_coplanar(*points):
        """

        This function tests whether passed points are coplanar or not.
        It uses the fact that the triple scalar product of three vectors
        vanishes if the vectors are coplanar. Which means that the volume
        of the solid described by them will have to be zero for coplanarity.

        Parameters
        ==========

        A set of points 3D points

        Returns
        =======

        boolean

        Examples
        ========

        >>> from diofant import Point3D
        >>> p1 = Point3D(1, 2, 2)
        >>> p2 = Point3D(2, 7, 2)
        >>> p3 = Point3D(0, 0, 2)
        >>> p4 = Point3D(1, 1, 2)
        >>> Point3D.are_coplanar(p1, p2, p3, p4)
        True
        >>> p5 = Point3D(0, 1, 3)
        >>> Point3D.are_coplanar(p1, p2, p3, p5)
        False
        """
        from .plane import Plane
        points = list(set(points))
        if len(points) < 3:
            raise ValueError('At least 3 points are needed to define a plane.')
        a, b = points[:2]
        for i, c in enumerate(points[2:]):
            try:
                p = Plane(a, b, c)
                for j in (0, 1, i):
                    points.pop(j)
                return all(p.is_coplanar(i) for i in points)
            except ValueError:
                pass
        raise ValueError('At least 3 non-collinear points needed to define plane.')

    def intersection(self, o):
        """The intersection between this point and another point.

        Parameters
        ==========

        other : Point

        Returns
        =======

        intersection : list of Points

        Notes
        =====

        The return value will either be an empty list if there is no
        intersection, otherwise it will contain this point.

        Examples
        ========

        >>> from diofant import Point3D
        >>> p1, p2, p3 = Point3D(0, 0, 0), Point3D(1, 1, 1), Point3D(0, 0, 0)
        >>> p1.intersection(p2)
        []
        >>> p1.intersection(p3)
        [Point3D(0, 0, 0)]

        """
        if isinstance(o, Point3D):
            if self == o:
                return [self]
            return []

        return o.intersection(self)

    def scale(self, x=1, y=1, z=1, pt=None):
        """Scale the coordinates of the Point by multiplying by
        ``x`` and ``y`` after subtracting ``pt`` -- default is (0, 0) --
        and then adding ``pt`` back again (i.e. ``pt`` is the point of
        reference for the scaling).

        See Also
        ========

        translate

        Examples
        ========

        >>> from diofant import Point3D
        >>> t = Point3D(1, 1, 1)
        >>> t.scale(2)
        Point3D(2, 1, 1)
        >>> t.scale(2, 2)
        Point3D(2, 2, 1)

        """
        if pt:
            pt = Point3D(pt)
            return self.translate(*(-pt).args).scale(x, y, z).translate(*pt.args)
        return Point3D(self.x*x, self.y*y, self.z*z)

    def translate(self, x=0, y=0, z=0):
        """Shift the Point by adding x and y to the coordinates of the Point.

        See Also
        ========

        diofant.geometry.entity.GeometryEntity.rotate
        diofant.geometry.entity.GeometryEntity.scale

        Examples
        ========

        >>> from diofant import Point3D
        >>> t = Point3D(0, 1, 1)
        >>> t.translate(2)
        Point3D(2, 1, 1)
        >>> t.translate(2, 2)
        Point3D(2, 3, 1)
        >>> t + Point3D(2, 2, 2)
        Point3D(2, 3, 3)
        """
        return Point3D(self.x + x, self.y + y, self.z + z)

    def transform(self, matrix):
        """Return the point after applying the transformation described
        by the 4x4 Matrix, ``matrix``.

        See Also
        ========

        diofant.geometry.entity.GeometryEntity.rotate
        diofant.geometry.entity.GeometryEntity.scale
        diofant.geometry.entity.GeometryEntity.translate
        """
        try:
            col, row = matrix.shape
            valid_matrix = matrix.is_square and col == 4
        except AttributeError:
            # We hit this block if matrix argument is not actually a Matrix.
            valid_matrix = False
        if not valid_matrix:
            raise ValueError("The argument to the transform function must be "
                             + "a 4x4 matrix")
        from ..matrices import Transpose
        x, y, z = self.args
        m = Transpose(matrix)
        return Point3D(*(Matrix(1, 4, [x, y, z, 1])*m).tolist()[0][:3])



from ..logic.boolalg import Boolean, BooleanAtom
from .basic import S
from .compatibility import ordered
from .evalf import EvalfMixin
from .evaluate import global_evaluate
from .expr import Expr
from .function import _coeff_isneg
from .symbol import Dummy, Symbol
from .sympify import sympify


__all__ = (
    'Rel', 'Eq', 'Ne', 'Lt', 'Le', 'Gt', 'Ge',
    'Relational', 'Equality', 'Unequality', 'StrictLessThan', 'LessThan',
    'StrictGreaterThan', 'GreaterThan',
)


# Note, see issue sympy/sympy#4986.  Ideally, we wouldn't want to subclass both Boolean
# and Expr.

class Relational(Boolean, Expr, EvalfMixin):
    """Base class for all relation types.

    Subclasses of Relational should generally be instantiated directly, but
    Relational can be instantiated with a valid `rop` value to dispatch to
    the appropriate subclass.

    Parameters
    ==========
    rop : str or None
        Indicates what subclass to instantiate.  Valid values can be found
        in the keys of Relational.ValidRelationalOperator.

    Examples
    ========

    >>> from diofant import Rel
    >>> from diofant.abc import x, y
    >>> Rel(y, x+x**2, '==')
    Eq(y, x**2 + x)

    """

    is_Relational = True

    # ValidRelationOperator - Defined below, because the necessary classes
    #   have not yet been defined

    def __new__(cls, lhs, rhs, rop=None, **assumptions):
        # If called by a subclass, do nothing special and pass on to Expr.
        if cls is not Relational:
            return Expr.__new__(cls, lhs, rhs, **assumptions)
        # If called directly with an operator, look up the subclass
        # corresponding to that operator and delegate to it
        try:
            cls = cls.ValidRelationOperator[rop]
            return cls(lhs, rhs, **assumptions)
        except KeyError:
            raise ValueError("Invalid relational operator symbol: %r" % rop)

    @property
    def lhs(self):
        """The left-hand side of the relation."""
        return self.args[0]

    @property
    def rhs(self):
        """The right-hand side of the relation."""
        return self.args[1]

    @property
    def reversed(self):
        """Return the relationship with sides (and sign) reversed.

        Examples
        ========

        >>> from diofant import Eq
        >>> from diofant.abc import x
        >>> Eq(x, 1)
        Eq(x, 1)
        >>> _.reversed
        Eq(1, x)
        >>> x < 1
        x < 1
        >>> _.reversed
        1 > x
        """
        ops = {Gt: Lt, Ge: Le, Lt: Gt, Le: Ge}
        a, b = self.args
        return ops.get(self.func, self.func)(b, a, evaluate=False)

    def _eval_evalf(self, prec):
        return self.func(*[s._evalf(prec) for s in self.args])

    @property
    def canonical(self):
        """Return a canonical form of the relational.

        The rules for the canonical form, in order of decreasing priority are:
            1) Number on right if left is not a Number;
            2) Symbol on the left;
            3) Gt/Ge changed to Lt/Le;
            4) Lt/Le are unchanged;
            5) Eq and Ne get ordered args.
        """
        r = self
        if r.func in (Ge, Gt):
            r = r.reversed
        elif r.func in (Lt, Le):
            pass
        elif r.func in (Eq, Ne):
            r = r.func(*ordered(r.args), evaluate=False)
        else:  # pragma: no cover
            raise NotImplemented
        if r.lhs.is_Number and not r.rhs.is_Number:
            r = r.reversed
        elif r.rhs.is_Symbol and not r.lhs.is_Symbol:
            r = r.reversed
        if _coeff_isneg(r.lhs):
            r = r.reversed.func(-r.lhs, -r.rhs, evaluate=False)
        return r

    def equals(self, other, failing_expression=False):
        """Return True if the sides of the relationship are mathematically
        identical and the type of relationship is the same.
        If failing_expression is True, return the expression whose truth value
        was unknown.
        """
        if isinstance(other, Relational):
            if self == other or self.reversed == other:
                return True
            a, b = self, other
            if a.func in (Eq, Ne) or b.func in (Eq, Ne):
                if a.func != b.func:
                    return False
                l = a.lhs.equals(b.lhs, failing_expression=failing_expression)
                r = a.rhs.equals(b.rhs, failing_expression=failing_expression)
                if l is True:
                    return r
                if r is True:
                    return l
                lr = a.lhs.equals(b.rhs, failing_expression=failing_expression)
                rl = a.rhs.equals(b.lhs, failing_expression=failing_expression)
                if lr is True:
                    return rl
                if rl is True:
                    return lr
                e = (l, r, lr, rl)
                if all(i is False for i in e):
                    return False
                if failing_expression:
                    return a.lhs - a.rhs - b.lhs + b.rhs
            else:
                if b.func != a.func:
                    b = b.reversed
                if a.func != b.func:
                    return False
                l = a.lhs.equals(b.lhs, failing_expression=failing_expression)
                if l is False:
                    return False
                r = a.rhs.equals(b.rhs, failing_expression=failing_expression)
                if r is False:
                    return False
                if l is True:
                    return r
                return l

    def _eval_simplify(self, ratio, measure):
        r = self.func(self.lhs.simplify(ratio=ratio, measure=measure),
                      self.rhs.simplify(ratio=ratio, measure=measure))

        if r.is_Relational:
            dif = r.lhs - r.rhs
            # We want a Number to compare with zero and be sure to get a
            # True/False answer.  Check if we can deduce that dif is
            # definitively zero or non-zero.  If non-zero, replace with an
            # approximation.  If .equals(0) gives None, cannot be deduced.
            if not dif.has(Dummy, Symbol):
                know = dif.equals(0)
                if know:
                    dif = S.Zero
                elif know is False:
                    if isinstance(r, Eq):
                        return False
                    elif isinstance(r, Ne):
                        return True
            # Can definitively compare a Number to zero, if appropriate.
            if dif.is_Number and dif.is_extended_real:
                # Always T/F (we never return an expression w/ the evalf)
                r = r.func._eval_relation(dif, S.Zero)

        r = r.canonical
        if measure(r) < ratio*measure(self):
            return r
        else:
            return self

    def __bool__(self):
        raise TypeError("cannot determine truth value of Relational")

    def as_set(self):
        """
        Rewrites univariate inequality in terms of real sets

        Examples
        ========

        >>> from diofant import Symbol, Eq
        >>> x = Symbol('x', extended_real=True)
        >>> (x>0).as_set()
        (0, oo)
        >>> Eq(x, 0).as_set()
        {0}

        """
        from ..solvers.inequalities import solve_univariate_inequality
        syms = self.free_symbols

        if len(syms) == 1:
            sym = syms.pop()
        else:  # pragma: no cover
            raise NotImplementedError("Sorry, Relational.as_set procedure"
                                      " is not yet implemented for"
                                      " multivariate expressions")

        return solve_univariate_inequality(self, sym, relational=False)


Rel = Relational


class Equality(Relational):
    """An equal relation between two objects.

    Represents that two objects are equal.  If they can be easily shown
    to be definitively equal (or unequal), this will reduce to True (or
    False).  Otherwise, the relation is maintained as an unevaluated
    Equality object.  Use the ``simplify`` function on this object for
    more nontrivial evaluation of the equality relation.

    As usual, the keyword argument ``evaluate=False`` can be used to
    prevent any evaluation.

    Examples
    ========

    >>> from diofant import Eq, simplify, exp, cos
    >>> from diofant.abc import x, y
    >>> Eq(y, x + x**2)
    Eq(y, x**2 + x)
    >>> Eq(2, 5)
    false
    >>> Eq(2, 5, evaluate=False)
    Eq(2, 5)
    >>> _.doit()
    false
    >>> Eq(exp(x), exp(x).rewrite(cos))
    Eq(E**x, sinh(x) + cosh(x))
    >>> simplify(_)
    true

    See Also
    ========

    diofant.logic.boolalg.Equivalent : for representing equality between two
        boolean expressions

    Notes
    =====

    This class is not the same as the == operator.  The == operator tests
    for exact structural equality between two expressions; this class
    compares expressions mathematically.

    If either object defines an `_eval_Eq` method, it can be used in place of
    the default algorithm.  If `lhs._eval_Eq(rhs)` or `rhs._eval_Eq(lhs)`
    returns anything other than None, that return value will be substituted for
    the Equality.  If None is returned by `_eval_Eq`, an Equality object will
    be created as usual.
    """

    rel_op = '=='

    is_Equality = True

    def __new__(cls, lhs, rhs=0, **options):
        lhs = sympify(lhs, strict=True)
        rhs = sympify(rhs, strict=True)

        evaluate = options.pop('evaluate', global_evaluate[0])

        if evaluate:
            # If one expression has an _eval_Eq, return its results.
            if hasattr(lhs, '_eval_Eq'):
                r = lhs._eval_Eq(rhs)
                if r is not None:
                    return r
            if hasattr(rhs, '_eval_Eq'):
                r = rhs._eval_Eq(lhs)
                if r is not None:
                    return r
            # If expressions have the same structure, they must be equal.
            if lhs == rhs:
                return S.true
            elif all(isinstance(i, BooleanAtom) for i in (rhs, lhs)):
                return S.false  # equal args already evaluated

            # If appropriate, check if the difference evaluates.  Detect
            # incompatibility such as lhs real and rhs not real.
            if isinstance(lhs, Expr) and isinstance(rhs, Expr):
                r = (lhs - rhs).is_zero
                if r is not None:
                    return sympify(r, strict=True)

        return Relational.__new__(cls, lhs, rhs, **options)


Eq = Equality


class Unequality(Relational):
    """An unequal relation between two objects.

    Represents that two objects are not equal.  If they can be shown to be
    definitively equal, this will reduce to False; if definitively unequal,
    this will reduce to True.  Otherwise, the relation is maintained as an
    Unequality object.

    Examples
    ========

    >>> from diofant import Ne
    >>> from diofant.abc import x, y
    >>> Ne(y, x+x**2)
    Ne(y, x**2 + x)

    See Also
    ========
    Equality

    Notes
    =====
    This class is not the same as the != operator.  The != operator tests
    for exact structural equality between two expressions; this class
    compares expressions mathematically.

    This class is effectively the inverse of Equality.  As such, it uses the
    same algorithms, including any available `_eval_Eq` methods.
    """

    rel_op = '!='

    def __new__(cls, lhs, rhs, **options):
        lhs = sympify(lhs, strict=True)
        rhs = sympify(rhs, strict=True)

        evaluate = options.pop('evaluate', global_evaluate[0])

        if evaluate:
            is_equal = Equality(lhs, rhs)
            if is_equal == S.true or is_equal == S.false:
                return ~is_equal

        return Relational.__new__(cls, lhs, rhs, **options)


Ne = Unequality


class _Inequality(Relational):
    """Internal base class for all *Than types.

    Each subclass must implement _eval_relation to provide the method for
    comparing two real numbers.
    """

    def __new__(cls, lhs, rhs, **options):
        lhs = sympify(lhs, strict=True)
        rhs = sympify(rhs, strict=True)

        evaluate = options.pop('evaluate', global_evaluate[0])

        if evaluate:
            # First we invoke the appropriate inequality method of `lhs`
            # (e.g., `lhs.__lt__`).  That method will try to reduce to
            # boolean or raise an exception.  It may keep calling
            # superclasses until it reaches `Expr` (e.g., `Expr.__lt__`).
            # In some cases, `Expr` will just invoke us again (if neither it
            # nor a subclass was able to reduce to boolean or raise an
            # exception).  In that case, it must call us with
            # `evaluate=False` to prevent infinite recursion.
            return cls._eval_relation(lhs, rhs)

        # make a "non-evaluated" Expr for the inequality
        return Relational.__new__(cls, lhs, rhs, **options)


class _Greater(_Inequality):
    """Not intended for general use

    _Greater is only used so that GreaterThan and StrictGreaterThan may subclass
    it for the .gts and .lts properties.
    """

    @property
    def gts(self):
        """Greater than side argument"""
        return self.args[0]

    @property
    def lts(self):
        """Less than side argument"""
        return self.args[1]


class _Less(_Inequality):
    """Not intended for general use.

    _Less is only used so that LessThan and StrictLessThan may subclass it for
    the .gts and .lts properties.
    """

    @property
    def gts(self):
        """Greater than side argument"""
        return self.args[1]

    @property
    def lts(self):
        """Less than side argument"""
        return self.args[0]


class GreaterThan(_Greater):
    r"""Class representations of inequalities.

    Extended Summary
    ================

    The ``*Than`` classes represent unequal relationships, where the left-hand
    side is generally bigger or smaller than the right-hand side.  For example,
    the GreaterThan class represents an unequal relationship where the
    left-hand side is at least as big as the right side, if not bigger.  In
    mathematical notation:

    lhs >= rhs

    In total, there are four ``*Than`` classes, to represent the four
    inequalities:

    +-----------------+--------+
    |Class Name       | Symbol |
    +=================+========+
    |GreaterThan      | (>=)   |
    +-----------------+--------+
    |LessThan         | (<=)   |
    +-----------------+--------+
    |StrictGreaterThan| (>)    |
    +-----------------+--------+
    |StrictLessThan   | (<)    |
    +-----------------+--------+

    All classes take two arguments, lhs and rhs.

    +----------------------------+-----------------+
    |Signature Example           | Math equivalent |
    +============================+=================+
    |GreaterThan(lhs, rhs)       |   lhs >= rhs    |
    +----------------------------+-----------------+
    |LessThan(lhs, rhs)          |   lhs <= rhs    |
    +----------------------------+-----------------+
    |StrictGreaterThan(lhs, rhs) |   lhs >  rhs    |
    +----------------------------+-----------------+
    |StrictLessThan(lhs, rhs)    |   lhs <  rhs    |
    +----------------------------+-----------------+

    In addition to the normal .lhs and .rhs of Relations, ``*Than`` inequality
    objects also have the .lts and .gts properties, which represent the "less
    than side" and "greater than side" of the operator.  Use of .lts and .gts
    in an algorithm rather than .lhs and .rhs as an assumption of inequality
    direction will make more explicit the intent of a certain section of code,
    and will make it similarly more robust to client code changes:

    >>> from diofant import GreaterThan, StrictGreaterThan
    >>> from diofant import LessThan,    StrictLessThan
    >>> from diofant import And, Ge, Gt, Le, Lt, Rel, S, Integer
    >>> from diofant.abc import x, y, z
    >>> from diofant.core.relational import Relational

    >>> e = GreaterThan(x, 1)
    >>> e
    x >= 1
    >>> '%s >= %s is the same as %s <= %s' % (e.gts, e.lts, e.lts, e.gts)
    'x >= 1 is the same as 1 <= x'

    Examples
    ========

    One generally does not instantiate these classes directly, but uses various
    convenience methods:

    >>> e1 = Ge(x, 2)  # Ge is a convenience wrapper
    >>> print(e1)
    x >= 2

    >>> rels = Ge(x, 2), Gt(x, 2), Le(x, 2), Lt(x, 2)
    >>> print('%s\n%s\n%s\n%s' % rels)
    x >= 2
    x > 2
    x <= 2
    x < 2

    Another option is to use the Python inequality operators (>=, >, <=, <)
    directly.  Their main advantage over the Ge, Gt, Le, and Lt counterparts, is
    that one can write a more "mathematical looking" statement rather than
    littering the math with oddball function calls.  However there are certain
    (minor) caveats of which to be aware (search for 'gotcha', below).

    >>> e2 = x >= 2
    >>> print(e2)
    x >= 2
    >>> print("e1: %s,    e2: %s" % (e1, e2))
    e1: x >= 2,    e2: x >= 2
    >>> e1 == e2
    True

    However, it is also perfectly valid to instantiate a ``*Than`` class less
    succinctly and less conveniently:

    >>> rels = Rel(x, 1, '>='), Relational(x, 1, '>='), GreaterThan(x, 1)
    >>> print('%s\n%s\n%s' % rels)
    x >= 1
    x >= 1
    x >= 1

    >>> rels = Rel(x, 1, '>'), Relational(x, 1, '>'), StrictGreaterThan(x, 1)
    >>> print('%s\n%s\n%s' % rels)
    x > 1
    x > 1
    x > 1

    >>> rels = Rel(x, 1, '<='), Relational(x, 1, '<='), LessThan(x, 1)
    >>> print("%s\n%s\n%s" % rels)
    x <= 1
    x <= 1
    x <= 1

    >>> rels = Rel(x, 1, '<'), Relational(x, 1, '<'), StrictLessThan(x, 1)
    >>> print('%s\n%s\n%s' % rels)
    x < 1
    x < 1
    x < 1

    Notes
    =====

    There are a couple of "gotchas" when using Python's operators.

    The first enters the mix when comparing against a literal number as the lhs
    argument.  Due to the order that Python decides to parse a statement, it may
    not immediately find two objects comparable.  For example, to evaluate the
    statement (1 < x), Python will first recognize the number 1 as a native
    number, and then that x is *not* a native number.  At this point, because a
    native Python number does not know how to compare itself with a Diofant object
    Python will try the reflective operation, (x > 1).  Unfortunately, there is
    no way available to Diofant to recognize this has happened, so the statement
    (1 < x) will turn silently into (x > 1).

    >>> e1 = x >  1
    >>> e2 = x >= 1
    >>> e3 = x <  1
    >>> e4 = x <= 1
    >>> e5 = 1 >  x
    >>> e6 = 1 >= x
    >>> e7 = 1 <  x
    >>> e8 = 1 <= x
    >>> print("%s     %s\n"*4 % (e1, e2, e3, e4, e5, e6, e7, e8))
    x > 1     x >= 1
    x < 1     x <= 1
    x < 1     x <= 1
    x > 1     x >= 1

    If the order of the statement is important (for visual output to the
    console, perhaps), one can work around this annoyance in a couple ways: (1)
    "sympify" the literal before comparison, (2) use one of the wrappers, or (3)
    use the less succinct methods described above:

    >>> e1 = Integer(1) >  x
    >>> e2 = Integer(1) >= x
    >>> e3 = Integer(1) <  x
    >>> e4 = Integer(1) <= x
    >>> e5 = Gt(1, x)
    >>> e6 = Ge(1, x)
    >>> e7 = Lt(1, x)
    >>> e8 = Le(1, x)
    >>> print("%s     %s\n"*4 % (e1, e2, e3, e4, e5, e6, e7, e8))
    1 > x     1 >= x
    1 < x     1 <= x
    1 > x     1 >= x
    1 < x     1 <= x

    The other gotcha is with chained inequalities.  Occasionally, one may be
    tempted to write statements like:

    >>> e = x < y < z
    Traceback (most recent call last):
    ...
    TypeError: symbolic boolean expression has no truth value.

    Due to an implementation detail or decision of Python [1]_, there is no way
    for Diofant to reliably create that as a chained inequality.  To create a
    chained inequality, the only method currently available is to make use of
    And:

    >>> e = And(x < y, y < z)
    >>> type(e)
    And
    >>> e
    And(x < y, y < z)

    Note that this is different than chaining an equality directly via use of
    parenthesis (this is currently an open bug in Diofant [2]_):

    >>> e = (x < y) < z
    >>> type(e)
    <class 'diofant.core.relational.StrictLessThan'>
    >>> e
    (x < y) < z

    Any code that explicitly relies on this latter functionality will not be
    robust as this behaviour is completely wrong and will be corrected at some
    point.  For the time being (circa Jan 2012), use And to create chained
    inequalities.

    References
    ==========

    .. [1] This implementation detail is that Python provides no reliable
       method to determine that a chained inequality is being built.  Chained
       comparison operators are evaluated pairwise, using "and" logic (see
       https://docs.python.org/3/reference/expressions.html#not-in).  This is done
       in an efficient way, so that each object being compared is only
       evaluated once and the comparison can short-circuit.  For example, ``1
       > 2 > 3`` is evaluated by Python as ``(1 > 2) and (2 > 3)``.  The
       ``and`` operator coerces each side into a bool, returning the object
       itself when it short-circuits.  The bool of the --Than operators
       will raise TypeError on purpose, because Diofant cannot determine the
       mathematical ordering of symbolic expressions.  Thus, if we were to
       compute ``x > y > z``, with ``x``, ``y``, and ``z`` being Symbols,
       Python converts the statement (roughly) into these steps:

        (1) x > y > z
        (2) (x > y) and (y > z)
        (3) (GreaterThanObject) and (y > z)
        (4) (GreaterThanObject.__bool__()) and (y > z)
        (5) TypeError

       Because of the "and" added at step 2, the statement gets turned into a
       weak ternary statement, and the first object's __bool__ method will
       raise TypeError.  Thus, creating a chained inequality is not possible.

           In Python, there is no way to override the ``and`` operator, or to
           control how it short circuits, so it is impossible to make something
           like ``x > y > z`` work.  There was a PEP to change this,
           :pep:`335`, but it was officially closed in March, 2012.

    .. [2] For more information, see these two bug reports:

       "Separate boolean and symbolic relationals"
       `Issue sympy/sympy#4986 <https://github.com/sympy/sympy/issues/4986>`_

       "It right 0 < x < 1 ?"
       `Issue sympy/sympy#6059 <https://github.com/sympy/sympy/issues/6059>`_
    """

    rel_op = '>='

    @classmethod
    def _eval_relation(cls, lhs, rhs):
        return sympify(lhs >= rhs, strict=True)


Ge = GreaterThan


class LessThan(_Less):
    __doc__ = GreaterThan.__doc__

    rel_op = '<='

    @classmethod
    def _eval_relation(cls, lhs, rhs):
        return sympify(lhs <= rhs, strict=True)


Le = LessThan


class StrictGreaterThan(_Greater):
    __doc__ = GreaterThan.__doc__

    rel_op = '>'

    @classmethod
    def _eval_relation(cls, lhs, rhs):
        return sympify(lhs > rhs, strict=True)


Gt = StrictGreaterThan


class StrictLessThan(_Less):
    __doc__ = GreaterThan.__doc__

    rel_op = '<'

    @classmethod
    def _eval_relation(cls, lhs, rhs):
        return sympify(lhs < rhs, strict=True)


Lt = StrictLessThan


# A class-specific (not object-specific) data item used for a minor speedup.  It
# is defined here, rather than directly in the class, because the classes that
# it references have not been defined until now (e.g. StrictLessThan).
Relational.ValidRelationOperator = {
    None: Equality,
    '==': Equality,
    'eq': Equality,
    '!=': Unequality,
    '<>': Unequality,
    'ne': Unequality,
    '>=': GreaterThan,
    'ge': GreaterThan,
    '<=': LessThan,
    'le': LessThan,
    '>': StrictGreaterThan,
    'gt': StrictGreaterThan,
    '<': StrictLessThan,
    'lt': StrictLessThan,
}



import pytest

from diofant import I, asin, cbrt, exp, false, log, simplify, sin, sqrt
from diofant.core import (Dummy, E, Float, GoldenRatio, Integer, Mod, Mul, Pow,
                          Rational, Symbol, Wild, nan, oo, pi, zoo)
from diofant.core.facts import InconsistentAssumptions


__all__ = ()


def test_symbol_unset():
    x = Symbol('x', extended_real=True, integer=True)
    assert x.is_extended_real is True
    assert x.is_integer is True
    assert x.is_imaginary is None
    assert x.is_noninteger is False
    assert x.is_number is False


def test_zero():
    z = Integer(0)
    assert z.is_commutative is True
    assert z.is_integer is True
    assert z.is_rational is True
    assert z.is_algebraic is True
    assert z.is_transcendental is False
    assert z.is_extended_real is True
    assert z.is_complex is True
    assert z.is_noninteger is False
    assert z.is_irrational is False
    assert z.is_imaginary is True
    assert z.is_positive is False
    assert z.is_negative is False
    assert z.is_nonpositive is True
    assert z.is_nonnegative is True
    assert z.is_even is True
    assert z.is_odd is False
    assert z.is_finite is True
    assert z.is_infinite is False
    assert z.is_comparable is True
    assert z.is_prime is False
    assert z.is_composite is False
    assert z.is_number is True


def test_one():
    z = Integer(1)
    assert z.is_commutative is True
    assert z.is_integer is True
    assert z.is_rational is True
    assert z.is_algebraic is True
    assert z.is_transcendental is False
    assert z.is_extended_real is True
    assert z.is_complex is True
    assert z.is_noninteger is False
    assert z.is_irrational is False
    assert z.is_imaginary is False
    assert z.is_positive is True
    assert z.is_negative is False
    assert z.is_nonpositive is False
    assert z.is_nonnegative is True
    assert z.is_even is False
    assert z.is_odd is True
    assert z.is_finite is True
    assert z.is_infinite is False
    assert z.is_comparable is True
    assert z.is_prime is False
    assert z.is_number is True
    assert z.is_composite is False  # issue sympy/sympy#8807


def test_negativeone():
    z = Integer(-1)
    assert z.is_commutative is True
    assert z.is_integer is True
    assert z.is_rational is True
    assert z.is_algebraic is True
    assert z.is_transcendental is False
    assert z.is_extended_real is True
    assert z.is_complex is True
    assert z.is_noninteger is False
    assert z.is_irrational is False
    assert z.is_imaginary is False
    assert z.is_positive is False
    assert z.is_negative is True
    assert z.is_nonpositive is True
    assert z.is_nonnegative is False
    assert z.is_even is False
    assert z.is_odd is True
    assert z.is_finite is True
    assert z.is_infinite is False
    assert z.is_comparable is True
    assert z.is_prime is False
    assert z.is_composite is False
    assert z.is_number is True


def test_infinity():
    assert oo.is_commutative is True
    assert oo.is_integer is False
    assert oo.is_rational is False
    assert oo.is_algebraic is False
    assert oo.is_transcendental is False
    assert oo.is_extended_real is True
    assert oo.is_complex is False
    assert oo.is_noninteger is False
    assert oo.is_irrational is False
    assert oo.is_imaginary is False
    assert oo.is_positive is True
    assert oo.is_negative is False
    assert oo.is_nonpositive is False
    assert oo.is_nonnegative is True
    assert oo.is_even is False
    assert oo.is_odd is False
    assert oo.is_finite is False
    assert oo.is_infinite is True
    assert oo.is_comparable is True
    assert oo.is_prime is False
    assert oo.is_composite is False
    assert oo.is_number is True


def test_neg_infinity():
    mm = -oo

    assert mm.is_commutative is True
    assert mm.is_integer is False
    assert mm.is_rational is False
    assert mm.is_algebraic is False
    assert mm.is_transcendental is False
    assert mm.is_extended_real is True
    assert mm.is_complex is False
    assert mm.is_noninteger is False
    assert mm.is_irrational is False
    assert mm.is_imaginary is False
    assert mm.is_positive is False
    assert mm.is_negative is True
    assert mm.is_nonpositive is True
    assert mm.is_nonnegative is False
    assert mm.is_even is False
    assert mm.is_odd is False
    assert mm.is_finite is False
    assert mm.is_infinite is True
    assert mm.is_comparable is True
    assert mm.is_prime is False
    assert mm.is_composite is False
    assert mm.is_number is True


def test_zoo():
    assert zoo.is_complex is False
    assert zoo.is_real is False
    assert zoo.is_prime is False
    assert zoo.is_infinite


def test_nan():
    assert nan.is_commutative is True
    assert nan.is_integer is False
    assert nan.is_rational is False
    assert nan.is_algebraic is False
    assert nan.is_transcendental is False
    assert nan.is_extended_real is None
    assert nan.is_complex is False
    assert nan.is_noninteger is False
    assert nan.is_irrational is False
    assert nan.is_imaginary is False
    assert nan.is_positive is None
    assert nan.is_negative is None
    assert nan.is_nonpositive is None
    assert nan.is_nonnegative is None
    assert nan.is_even is False
    assert nan.is_odd is False
    assert nan.is_finite is False
    assert nan.is_infinite is None
    assert nan.is_comparable is False
    assert nan.is_prime is False
    assert nan.is_composite is False
    assert nan.is_number is True


def test_pos_rational():
    r = Rational(3, 4)
    assert r.is_commutative is True
    assert r.is_integer is False
    assert r.is_rational is True
    assert r.is_algebraic is True
    assert r.is_transcendental is False
    assert r.is_extended_real is True
    assert r.is_complex is True
    assert r.is_noninteger is True
    assert r.is_irrational is False
    assert r.is_imaginary is False
    assert r.is_positive is True
    assert r.is_negative is False
    assert r.is_nonpositive is False
    assert r.is_nonnegative is True
    assert r.is_even is False
    assert r.is_odd is False
    assert r.is_finite is True
    assert r.is_infinite is False
    assert r.is_comparable is True
    assert r.is_prime is False
    assert r.is_composite is False

    r = Rational(1, 4)
    assert r.is_nonpositive is False
    assert r.is_positive is True
    assert r.is_negative is False
    assert r.is_nonnegative is True
    r = Rational(5, 4)
    assert r.is_negative is False
    assert r.is_positive is True
    assert r.is_nonpositive is False
    assert r.is_nonnegative is True
    r = Rational(5, 3)
    assert r.is_nonnegative is True
    assert r.is_positive is True
    assert r.is_negative is False
    assert r.is_nonpositive is False


def test_neg_rational():
    r = Rational(-3, 4)
    assert r.is_positive is False
    assert r.is_nonpositive is True
    assert r.is_negative is True
    assert r.is_nonnegative is False
    r = Rational(-1, 4)
    assert r.is_nonpositive is True
    assert r.is_positive is False
    assert r.is_negative is True
    assert r.is_nonnegative is False
    r = Rational(-5, 4)
    assert r.is_negative is True
    assert r.is_positive is False
    assert r.is_nonpositive is True
    assert r.is_nonnegative is False
    r = Rational(-5, 3)
    assert r.is_nonnegative is False
    assert r.is_positive is False
    assert r.is_negative is True
    assert r.is_nonpositive is True


def test_pi():
    z = pi
    assert z.is_commutative is True
    assert z.is_integer is False
    assert z.is_rational is False
    assert z.is_algebraic is False
    assert z.is_transcendental is True
    assert z.is_extended_real is True
    assert z.is_complex is True
    assert z.is_noninteger is True
    assert z.is_irrational is True
    assert z.is_imaginary is False
    assert z.is_positive is True
    assert z.is_negative is False
    assert z.is_nonpositive is False
    assert z.is_nonnegative is True
    assert z.is_even is False
    assert z.is_odd is False
    assert z.is_finite is True
    assert z.is_infinite is False
    assert z.is_comparable is True
    assert z.is_prime is False
    assert z.is_composite is False


def test_E():
    z = E
    assert z.is_commutative is True
    assert z.is_integer is False
    assert z.is_rational is False
    assert z.is_algebraic is False
    assert z.is_transcendental is True
    assert z.is_extended_real is True
    assert z.is_complex is True
    assert z.is_noninteger is True
    assert z.is_irrational is True
    assert z.is_imaginary is False
    assert z.is_positive is True
    assert z.is_negative is False
    assert z.is_nonpositive is False
    assert z.is_nonnegative is True
    assert z.is_even is False
    assert z.is_odd is False
    assert z.is_finite is True
    assert z.is_infinite is False
    assert z.is_comparable is True
    assert z.is_prime is False
    assert z.is_composite is False


def test_I():
    z = I
    assert z.is_commutative is True
    assert z.is_integer is False
    assert z.is_rational is False
    assert z.is_algebraic is True
    assert z.is_transcendental is False
    assert z.is_extended_real is False
    assert z.is_complex is True
    assert z.is_noninteger is False
    assert z.is_irrational is False
    assert z.is_imaginary is True
    assert z.is_positive is False
    assert z.is_negative is False
    assert z.is_nonpositive is False
    assert z.is_nonnegative is False
    assert z.is_even is False
    assert z.is_odd is False
    assert z.is_finite is True
    assert z.is_infinite is False
    assert z.is_comparable is False
    assert z.is_prime is False
    assert z.is_composite is False


def test_symbol_real():
    # issue sympy/sympy#3848
    a = Symbol('a', extended_real=False)

    assert a.is_extended_real is False
    assert a.is_integer is False
    assert a.is_negative is False
    assert a.is_positive is False
    assert a.is_nonnegative is False
    assert a.is_nonpositive is False
    assert a.is_zero is False


def test_symbol_zero():
    x = Symbol('x', zero=True)
    assert x.is_positive is False
    assert x.is_nonpositive
    assert x.is_negative is False
    assert x.is_nonnegative
    assert x.is_zero is True
    assert x.is_nonzero is False
    assert x.is_finite is True


def test_symbol_positive():
    x = Symbol('x', positive=True)
    assert x.is_positive is True
    assert x.is_nonpositive is False
    assert x.is_negative is False
    assert x.is_nonnegative is True
    assert x.is_zero is False
    assert x.is_nonzero is True


def test_neg_symbol_positive():
    x = -Symbol('x', positive=True)
    assert x.is_positive is False
    assert x.is_nonpositive is True
    assert x.is_negative is True
    assert x.is_nonnegative is False
    assert x.is_zero is False
    assert x.is_nonzero is True


def test_symbol_nonpositive():
    x = Symbol('x', nonpositive=True)
    assert x.is_positive is False
    assert x.is_nonpositive is True
    assert x.is_negative is None
    assert x.is_nonnegative is None
    assert x.is_zero is None
    assert x.is_nonzero is None


def test_neg_symbol_nonpositive():
    x = -Symbol('x', nonpositive=True)
    assert x.is_positive is None
    assert x.is_nonpositive is None
    assert x.is_negative is False
    assert x.is_nonnegative is True
    assert x.is_zero is None
    assert x.is_nonzero is None


def test_symbol_falsepositive():
    x = Symbol('x', positive=False)
    assert x.is_positive is False
    assert x.is_nonpositive is None
    assert x.is_negative is None
    assert x.is_nonnegative is None
    assert x.is_zero is None
    assert x.is_nonzero is None


def test_neg_symbol_falsepositive():
    x = -Symbol('x', positive=False)
    assert x.is_positive is None
    assert x.is_nonpositive is None
    assert x.is_negative is False
    assert x.is_nonnegative is None
    assert x.is_zero is None
    assert x.is_nonzero is None


def test_symbol_falsepositive_real():
    x = Symbol('x', positive=False, extended_real=True)
    assert x.is_positive is False
    assert x.is_nonpositive is True
    assert x.is_negative is None
    assert x.is_nonnegative is None
    assert x.is_zero is None
    assert x.is_nonzero is None


def test_neg_symbol_falsepositive_real():
    x = -Symbol('x', positive=False, extended_real=True)
    assert x.is_positive is None
    assert x.is_nonpositive is None
    assert x.is_negative is False
    assert x.is_nonnegative is True
    assert x.is_zero is None
    assert x.is_nonzero is None


def test_symbol_falsenonnegative():
    x = Symbol('x', nonnegative=False)
    assert x.is_positive is False
    assert x.is_nonpositive is None
    assert x.is_negative is None
    assert x.is_nonnegative is False
    assert x.is_zero is False
    assert x.is_nonzero is True


@pytest.mark.xfail
def test_neg_symbol_falsenonnegative():
    x = -Symbol('x', nonnegative=False)
    assert x.is_positive is None
    assert x.is_nonpositive is False  # this currently returns None
    assert x.is_negative is False  # this currently returns None
    assert x.is_nonnegative is None
    assert x.is_zero is False  # this currently returns None
    assert x.is_nonzero is True  # this currently returns None


def test_symbol_falsenonnegative_real():
    x = Symbol('x', nonnegative=False, extended_real=True)
    assert x.is_positive is False
    assert x.is_nonpositive is True
    assert x.is_negative is True
    assert x.is_nonnegative is False
    assert x.is_zero is False
    assert x.is_nonzero is True


def test_neg_symbol_falsenonnegative_real():
    x = -Symbol('x', nonnegative=False, extended_real=True)
    assert x.is_positive is True
    assert x.is_nonpositive is False
    assert x.is_negative is False
    assert x.is_nonnegative is True
    assert x.is_zero is False
    assert x.is_nonzero is True


def test_prime():
    assert Integer(-1).is_prime is False
    assert Integer(-2).is_prime is False
    assert Integer(-4).is_prime is False
    assert Integer(0).is_prime is False
    assert Integer(1).is_prime is False
    assert Integer(2).is_prime is True
    assert Integer(17).is_prime is True
    assert Integer(4).is_prime is False


def test_composite():
    assert Integer(-1).is_composite is False
    assert Integer(-2).is_composite is False
    assert Integer(-4).is_composite is False
    assert Integer(0).is_composite is False
    assert Integer(2).is_composite is False
    assert Integer(17).is_composite is False
    assert Integer(4).is_composite is True


def test_prime_symbol():
    x = Symbol('x', prime=True)
    assert x.is_prime is True
    assert x.is_integer is True
    assert x.is_positive is True
    assert x.is_negative is False
    assert x.is_nonpositive is False
    assert x.is_nonnegative is True

    x = Symbol('x', prime=False)
    assert x.is_prime is False
    assert x.is_integer is None
    assert x.is_positive is None
    assert x.is_negative is None
    assert x.is_nonpositive is None
    assert x.is_nonnegative is None


def test_symbol_noncommutative():
    x = Symbol('x', commutative=True)
    assert x.is_complex is None

    x = Symbol('x', commutative=False)
    assert x.is_integer is False
    assert x.is_rational is False
    assert x.is_algebraic is False
    assert x.is_irrational is False
    assert x.is_extended_real is False
    assert x.is_complex is False


def test_other_symbol():
    x = Symbol('x', integer=True)
    assert x.is_integer is True
    assert x.is_extended_real is True

    x = Symbol('x', integer=True, nonnegative=True)
    assert x.is_integer is True
    assert x.is_nonnegative is True
    assert x.is_negative is False
    assert x.is_positive is None

    x = Symbol('x', integer=True, nonpositive=True)
    assert x.is_integer is True
    assert x.is_nonpositive is True
    assert x.is_positive is False
    assert x.is_negative is None

    x = Symbol('x', odd=True)
    assert x.is_odd is True
    assert x.is_even is False
    assert x.is_integer is True

    x = Symbol('x', odd=False)
    assert x.is_odd is False
    assert x.is_even is None
    assert x.is_integer is None

    x = Symbol('x', even=True)
    assert x.is_even is True
    assert x.is_odd is False
    assert x.is_integer is True

    x = Symbol('x', even=False)
    assert x.is_even is False
    assert x.is_odd is None
    assert x.is_integer is None

    x = Symbol('x', integer=True, nonnegative=True)
    assert x.is_integer is True
    assert x.is_nonnegative is True

    x = Symbol('x', integer=True, nonpositive=True)
    assert x.is_integer is True
    assert x.is_nonpositive is True

    with pytest.raises(AttributeError):
        x.is_extended_real = False

    x = Symbol('x', algebraic=True)
    assert x.is_transcendental is False
    x = Symbol('x', transcendental=True)
    assert x.is_algebraic is False
    assert x.is_rational is False
    assert x.is_integer is False


def test_sympyissue_3825():
    """catch: hash instability"""
    x = Symbol("x")
    y = Symbol("y")
    a1 = x + y
    a2 = y + x
    a2.is_comparable

    h1 = hash(a1)
    h2 = hash(a2)
    assert h1 == h2


def test_sympyissue_4822():
    z = cbrt(-1)*(1 - I*sqrt(3))
    assert z.is_extended_real in [True, None]


def test_hash_vs_typeinfo():
    """seemingly different typeinfo, but in fact equal"""

    # the following two are semantically equal
    x1 = Symbol('x', even=True)
    x2 = Symbol('x', integer=True, odd=False)

    assert hash(x1) == hash(x2)
    assert x1 == x2


def test_hash_vs_typeinfo_2():
    """different typeinfo should mean !eq"""
    # the following two are semantically different
    x = Symbol('x')
    x1 = Symbol('x', even=True)

    assert x != x1
    assert hash(x) != hash(x1)  # This might fail with very low probability


def test_hash_vs_eq():
    """catch: different hash for equal objects"""
    a = 1 + pi    # important: do not fold it into a Number instance
    ha = hash(a)  # it should be Add/Mul/... to trigger the bug

    a.is_positive   # this uses .evalf() and deduces it is positive
    assert a.is_positive is True

    # be sure that hash stayed the same
    assert ha == hash(a)

    # now b should be the same expression
    b = a.expand(trig=True)
    hb = hash(b)

    assert a == b
    assert ha == hb


def test_Add_is_pos_neg():
    # these cover lines not covered by the rest of tests in core
    n = Symbol('n', negative=True, infinite=True)
    nn = Symbol('n', nonnegative=True, infinite=True)
    np = Symbol('n', nonpositive=True, infinite=True)
    p = Symbol('p', positive=True, infinite=True)
    r = Dummy(extended_real=True, finite=False)
    x = Symbol('x')
    xf = Symbol('xb', finite=True, real=True)
    assert (n + p).is_positive is None
    assert (n + x).is_positive is None
    assert (p + x).is_positive is None
    assert (n + p).is_negative is None
    assert (n + x).is_negative is None
    assert (p + x).is_negative is None

    assert (n + xf).is_positive is False
    assert (p + xf).is_positive is True
    assert (n + xf).is_negative is True
    assert (p + xf).is_negative is False

    assert (x - oo).is_negative is None  # issue sympy/sympy#7798
    # issue sympy/sympy#8046, 16.2
    assert (p + nn).is_positive
    assert (n + np).is_negative
    assert (p + r).is_positive is None


def test_Add_is_imaginary():
    nn = Dummy(nonnegative=True, finite=True)
    assert (I*nn + I).is_imaginary  # issue sympy/sympy#8046, 17


def test_Add_is_algebraic():
    a = Symbol('a', algebraic=True)
    b = Symbol('a', algebraic=True)
    na = Symbol('na', algebraic=False)
    nb = Symbol('nb', algebraic=False)
    x = Symbol('x')
    assert (a + b).is_algebraic
    assert (na + nb).is_algebraic is None
    assert (a + na).is_algebraic is False
    assert (a + x).is_algebraic is None
    assert (na + x).is_algebraic is None


def test_Mul_is_algebraic():
    a = Symbol('a', algebraic=True)
    b = Symbol('a', algebraic=True)
    na = Symbol('na', algebraic=False)
    an = Symbol('an', algebraic=True, nonzero=True)
    nb = Symbol('nb', algebraic=False)
    x = Symbol('x')
    assert (a*b).is_algebraic
    assert (na*nb).is_algebraic is None
    assert (a*na).is_algebraic is None
    assert (an*na).is_algebraic is False
    assert (a*x).is_algebraic is None
    assert (na*x).is_algebraic is None


def test_Pow_is_algebraic():
    e = Symbol('e', algebraic=True)

    assert Pow(1, e, evaluate=False).is_algebraic
    assert Pow(0, e, evaluate=False).is_algebraic

    a = Symbol('a', algebraic=True)
    na = Symbol('na', algebraic=False)
    ia = Symbol('ia', algebraic=True, irrational=True)
    ib = Symbol('ib', algebraic=True, irrational=True)
    r = Symbol('r', rational=True, nonzero=True)
    x = Symbol('x')
    assert (a**r).is_algebraic
    assert (a**x).is_algebraic is None
    assert (na**r).is_algebraic is False
    assert (ia**r).is_algebraic
    assert (ia**ib).is_algebraic is False

    assert (a**e).is_algebraic is None

    # Gelfond-Schneider constant:
    assert Pow(2, sqrt(2), evaluate=False).is_algebraic is False

    assert Pow(GoldenRatio, sqrt(3), evaluate=False).is_algebraic is False

    # sympy/sympy#8649
    t = Symbol('t', real=True, transcendental=True)
    n = Symbol('n', integer=True)
    assert (t**n).is_algebraic is None
    assert (t**n).is_integer is None


def test_Mul_is_infinite():
    x = Symbol('x')
    f = Symbol('f', finite=True)
    i = Symbol('i', infinite=True)
    z = Dummy(zero=True)
    nzf = Dummy(finite=True, zero=False)
    assert (x*f).is_finite is None
    assert (x*i).is_finite is None
    assert (f*i).is_finite is False
    assert (x*f*i).is_finite is None
    assert (z*i).is_finite is False
    assert (nzf*i).is_finite is False
    assert (z*f).is_finite is True
    assert Mul(0, f, evaluate=False).is_finite is True
    assert Mul(0, i, evaluate=False).is_finite is False

    assert (x*f).is_infinite is None
    assert (x*i).is_infinite is None
    assert (f*i).is_infinite is None
    assert (x*f*i).is_infinite is None
    assert (z*i).is_infinite is nan.is_infinite
    assert (nzf*i).is_infinite is True
    assert (z*f).is_infinite is False
    assert Mul(0, f, evaluate=False).is_infinite is False
    assert Mul(0, i, evaluate=False).is_infinite is nan.is_infinite


def test_special_is_rational():
    i = Symbol('i', integer=True)
    i2 = Symbol('i2', integer=True)
    ni = Symbol('ni', integer=True, nonzero=True)
    r = Symbol('r', rational=True)
    rn = Symbol('r', rational=True, nonzero=True)
    nr = Symbol('nr', irrational=True)
    x = Symbol('x')
    assert sqrt(3).is_rational is False
    assert (3 + sqrt(3)).is_rational is False
    assert (3*sqrt(3)).is_rational is False
    z = Symbol('z', zero=True)
    assert exp(z).is_rational
    assert exp(0, evaluate=False).is_rational
    assert exp(3).is_rational is False
    assert exp(ni).is_rational is False
    assert exp(rn).is_rational is False
    assert exp(x).is_rational is None
    assert exp(log(3), evaluate=False).is_rational is True
    assert log(exp(3), evaluate=False).is_rational is True
    assert log(3).is_rational is False
    assert log(ni + 1).is_rational is False
    assert log(rn + 1).is_rational is False
    assert log(x).is_rational is None
    assert (sqrt(3) + sqrt(5)).is_rational is None
    assert (sqrt(3) + pi).is_rational is False
    assert (x**i).is_rational is None
    assert (i**i).is_rational is True
    assert (i**i2).is_rational is None
    assert (r**i).is_rational is None
    assert (r**r).is_rational is None
    assert (r**x).is_rational is None
    assert (nr**i).is_rational is None  # issue sympy/sympy#8598
    assert (nr**Symbol('z', zero=True)).is_rational
    assert sin(1).is_rational is False
    assert sin(ni).is_rational is False
    assert sin(rn).is_rational is False
    assert sin(x).is_rational is None
    assert asin(rn).is_rational is False
    assert sin(asin(3), evaluate=False).is_rational is True


@pytest.mark.xfail
def test_sympyissue_6275():
    x = Symbol('x')
    # both zero or both Muls...but neither "change would be very appreciated.
    # This is similar to x/x => 1 even though if x = 0, it is really nan.
    assert isinstance(x*0, type(0*oo))
    if 0*oo is nan:
        b = Symbol('b', finite=None)
        assert (b*0).is_zero is None


def test_sanitize_assumptions():
    # issue sympy/sympy#6666
    for cls in (Symbol, Dummy, Wild):
        x = cls('x', extended_real=1, positive=0)
        assert x.is_extended_real is True
        assert x.is_positive is False
        assert cls('', extended_real=True, positive=None).is_positive is None
        pytest.raises(ValueError, lambda: cls('', commutative=None))
    pytest.raises(ValueError, lambda: Symbol._sanitize({'commutative': None}))


def test_special_assumptions():
    e = -3 - sqrt(5) + (-sqrt(10)/2 - sqrt(2)/2)**2
    assert simplify(e < 0) is false
    assert simplify(e > 0) is false
    assert (e == 0) is False  # it's not a literal 0
    assert e.equals(0) is True


def test_inconsistent():
    # cf. issues sympy/sympy#5795 and sympy/sympy#5545
    pytest.raises(InconsistentAssumptions, lambda: Symbol('x', extended_real=True,
                                                          commutative=False))


def test_sympyissue_6631():
    assert ((-1)**(I)).is_extended_real is True
    assert ((-1)**(I*2)).is_extended_real is True
    assert ((-1)**(I/2)).is_extended_real is True
    assert ((-1)**(I*pi)).is_extended_real is True
    assert (I**(I + 2)).is_extended_real is True


def test_sympyissue_2730():
    assert (1/(1 + I)).is_extended_real is False


def test_sympyissue_4149():
    assert (3 + I).is_complex
    assert (3 + I).is_imaginary is False
    assert (3*I + pi*I).is_imaginary
    y = Symbol('y', real=True)
    assert (3*I + pi*I + y*I).is_imaginary is True
    p = Symbol('p', positive=True, finite=True)
    assert (3*I + pi*I + p*I).is_imaginary
    n = Symbol('n', negative=True, finite=True)
    assert (-3*I - pi*I + n*I).is_imaginary

    i = Symbol('i', imaginary=True)
    assert ([(i**a).is_imaginary for a in range(4)] ==
            [False, True, False, True])

    # tests from the PR sympy/sympy#7887:
    e = -sqrt(3)*I/2 + Float(0.866025403784439)*I
    assert e.is_extended_real is False
    assert e.is_imaginary


def test_sympyissue_2920():
    n = Symbol('n', real=True, negative=True)
    assert sqrt(n).is_imaginary


def test_sympyissue_7899():
    x = Symbol('x', extended_real=True)
    assert (I*x).is_extended_real is None
    assert ((x - I)*(x - 1)).is_zero is None
    assert ((x - I)*(x - 1)).is_extended_real is None


@pytest.mark.xfail
def test_sympyissue_7993():
    x = Dummy(integer=True)
    y = Dummy(noninteger=True)
    assert (x - y).is_zero is False


def test_sympyissue_8075():
    pytest.raises(InconsistentAssumptions, lambda: Dummy(zero=True, finite=False))
    pytest.raises(InconsistentAssumptions, lambda: Dummy(zero=True, infinite=True))


def test_sympyissue_8642():
    x = Symbol('x', extended_real=True, integer=False)
    assert (x*2).is_integer is None


def test_sympyissue_9165():
    z = Symbol('z', zero=True)
    f = Symbol('f', finite=False)
    assert 0/z == nan
    assert 0*(1/z) == nan
    assert 0*f == nan


def test_sympyissue_10024():
    x = Dummy('x')
    assert Mod(x, 2*pi).is_zero is None



from diofant.core import Eq, Rational, Symbol
from diofant.functions import Piecewise, cos, cot, csc, log, sec, sin, tan
from diofant.integrals.trigonometry import trigintegrate


__all__ = ()

x = Symbol('x')


def test_trigintegrate_odd():
    assert trigintegrate(Rational(1), x) == x
    assert trigintegrate(x, x) is None
    assert trigintegrate(x**2, x) is None

    assert trigintegrate(sin(x), x) == -cos(x)
    assert trigintegrate(cos(x), x) == sin(x)

    assert trigintegrate(sin(3*x), x) == -cos(3*x)/3
    assert trigintegrate(cos(3*x), x) == sin(3*x)/3

    y = Symbol('y')
    assert trigintegrate(sin(y*x), x) == \
        Piecewise((0, Eq(y, 0)), (-cos(y*x)/y, True))
    assert trigintegrate(cos(y*x), x) == \
        Piecewise((x, Eq(y, 0)), (sin(y*x)/y, True))
    assert trigintegrate(sin(y*x)**2, x) == \
        Piecewise((0, Eq(y, 0)), ((x*y/2 - sin(x*y)*cos(x*y)/2)/y, True))
    assert trigintegrate(sin(y*x)*cos(y*x), x) == \
        Piecewise((0, Eq(y, 0)), (sin(x*y)**2/(2*y), True))
    assert trigintegrate(cos(y*x)**2, x) == \
        Piecewise((x, Eq(y, 0)), ((x*y/2 + sin(x*y)*cos(x*y)/2)/y, True))

    y = Symbol('y', positive=True)
    # TODO: remove conds='none' below. For this to work we would have to rule
    #       out (e.g. by trying solve) the condition y = 0, incompatible with
    #       y.is_positive being True.
    assert trigintegrate(sin(y*x), x, conds='none') == -cos(y*x)/y
    assert trigintegrate(cos(y*x), x, conds='none') == sin(y*x)/y

    assert trigintegrate(sin(x)*cos(x), x) == sin(x)**2/2
    assert trigintegrate(sin(x)*cos(x)**2, x) == -cos(x)**3/3
    assert trigintegrate(sin(x)**2*cos(x), x) == sin(x)**3/3

    # check if it selects right function to substitute,
    # so the result is kept simple
    assert trigintegrate(sin(x)**7 * cos(x), x) == sin(x)**8/8
    assert trigintegrate(sin(x) * cos(x)**7, x) == -cos(x)**8/8

    assert trigintegrate(sin(x)**7 * cos(x)**3, x) == \
        -sin(x)**10/10 + sin(x)**8/8
    assert trigintegrate(sin(x)**3 * cos(x)**7, x) == \
        cos(x)**10/10 - cos(x)**8/8


def test_trigintegrate_even():
    assert trigintegrate(sin(x)**2, x) == x/2 - cos(x)*sin(x)/2
    assert trigintegrate(cos(x)**2, x) == x/2 + cos(x)*sin(x)/2

    assert trigintegrate(sin(3*x)**2, x) == x/2 - cos(3*x)*sin(3*x)/6
    assert trigintegrate(cos(3*x)**2, x) == x/2 + cos(3*x)*sin(3*x)/6
    assert trigintegrate(sin(x)**2 * cos(x)**2, x) == \
        x/8 - sin(2*x)*cos(2*x)/16

    assert trigintegrate(sin(x)**4 * cos(x)**2, x) == \
        x/16 - sin(x) * cos(x)/16 - sin(x)**3*cos(x)/24 + \
        sin(x)**5*cos(x)/6

    assert trigintegrate(sin(x)**2 * cos(x)**4, x) == \
        x/16 + cos(x) * sin(x)/16 + cos(x)**3*sin(x)/24 - \
        cos(x)**5*sin(x)/6

    assert trigintegrate(sin(x)**(-4), x) == -2*cos(x)/(3*sin(x)) \
        - cos(x)/(3*sin(x)**3)

    assert trigintegrate(cos(x)**(-6), x) == sin(x)/(5*cos(x)**5) \
        + 4*sin(x)/(15*cos(x)**3) + 8*sin(x)/(15*cos(x))


def test_trigintegrate_mixed():
    assert trigintegrate(sin(x)*sec(x), x) == -log(sin(x)**2 - 1)/2
    assert trigintegrate(sin(x)*csc(x), x) == x
    assert trigintegrate(sin(x)*cot(x), x) == sin(x)

    assert trigintegrate(cos(x)*sec(x), x) == x
    assert trigintegrate(cos(x)*csc(x), x) == log(cos(x)**2 - 1)/2
    assert trigintegrate(cos(x)*tan(x), x) == -cos(x)
    assert trigintegrate(cos(x)*cot(x), x) == log(cos(x) - 1)/2 \
        - log(cos(x) + 1)/2 + cos(x)


def test_trigintegrate_symbolic():
    n = Symbol('n', integer=True)
    assert trigintegrate(cos(x)**n, x) is None
    assert trigintegrate(sin(x)**n, x) is None
    assert trigintegrate(cot(x)**n, x) is None



"""
Package for symbolic vector algebra in 3D.
"""

from .vector import Vector, VectorAdd, VectorMul, BaseVector, VectorZero  # noqa: F401
from .dyadic import Dyadic, DyadicAdd, DyadicMul, BaseDyadic, DyadicZero  # noqa: F401
from .scalar import BaseScalar  # noqa: F401
from .deloperator import Del  # noqa: F401
from .coordsysrect import CoordSysCartesian  # noqa: F401
from .functions import (express, matrix_to_vector, curl, divergence,  # noqa: F401
                        gradient, is_conservative, is_solenoidal,
                        scalar_potential, scalar_potential_difference)
from .point import Point  # noqa: F401
from .orienters import (AxisOrienter, BodyOrienter, SpaceOrienter,  # noqa: F401
                        QuaternionOrienter)



"""This tests diofant/core/basic.py with (ideally) no reference to subclasses
of Basic or Atom.
"""

import collections

import pytest

from diofant import Lambda, cos, sin
from diofant.abc import w, x, y, z
from diofant.core.basic import Atom, Basic, preorder_traversal
from diofant.core.compatibility import default_sort_key
from diofant.core.function import Function
from diofant.core.singleton import SingletonWithManagedProperties as Singleton
from diofant.core.singleton import S
from diofant.integrals.integrals import Integral


__all__ = ()


b1 = Basic()
b2 = Basic(b1)
b3 = Basic(b2)
b21 = Basic(b2, b1)


def test_structure():
    assert b21.args == (b2, b1)
    assert b21.func(*b21.args) == b21
    assert bool(b1)


def test_equality():
    instances = [b1, b2, b3, b21, Basic(b1, b1, b1), Basic]
    for i, b_i in enumerate(instances):
        for j, b_j in enumerate(instances):
            assert (b_i == b_j) == (i == j)
            assert (b_i != b_j) == (i != j)

    assert Basic() != []
    assert not(Basic() == [])
    assert Basic() != 0
    assert not(Basic() == 0)


def test_matches_basic():
    instances = [Basic(b1, b1, b2), Basic(b1, b2, b1), Basic(b2, b1, b1),
                 Basic(b1, b2), Basic(b2, b1), b2, b1]
    for i, b_i in enumerate(instances):
        for j, b_j in enumerate(instances):
            if i == j:
                assert b_j.match(b_i) == {}
            else:
                assert b_j.match(b_i) is None
    assert b1.match(b1) == {}


def test_has():
    assert b21.has(b1)
    assert b21.has(b3, b1)
    assert b21.has(Basic)
    assert not b1.has(b21, b3)
    assert not b21.has()


def test_subs():
    assert b21.subs(b2, b1) == Basic(b1, b1)
    assert b21.subs(b2, b21) == Basic(b21, b1)
    assert b3.subs(b2, b1) == b2

    assert b21.subs([(b2, b1), (b1, b2)]) == Basic(b2, b2)

    assert b21.subs({b1: b2, b2: b1}) == Basic(b2, b2)

    pytest.raises(ValueError, lambda: b21.subs('bad arg'))
    pytest.raises(ValueError, lambda: b21.subs(b1, b2, b3))

    assert b21.subs(collections.ChainMap({b1: b2}, {b2: b1})) == Basic(b2, b2)
    assert b21.subs(collections.OrderedDict([(b2, b1), (b1, b2)])) == Basic(b2, b2)


def test_rewrite():
    assert sin(1).rewrite() == sin(1)


def test_atoms():
    assert b21.atoms() == set()


def test_free_symbols_empty():
    assert b21.free_symbols == set()


def test_doit():
    assert b21.doit() == b21
    assert b21.doit(deep=False) == b21


def test_S():
    assert repr(S) == 'S'


def test_xreplace():
    assert b21.xreplace({b2: b1}) == Basic(b1, b1)
    assert b21.xreplace({b2: b21}) == Basic(b21, b1)
    assert b3.xreplace({b2: b1}) == b2
    assert Basic(b1, b2).xreplace({b1: b2, b2: b1}) == Basic(b2, b1)
    assert Atom(b1).xreplace({b1: b2}) == Atom(b1)
    assert Atom(b1).xreplace({Atom(b1): b2}) == b2
    pytest.raises(TypeError, lambda: b1.xreplace())
    pytest.raises(TypeError, lambda: b1.xreplace([b1, b2]))


def test_Singleton():
    global instantiated
    instantiated = 0

    class MyNewSingleton(Basic, metaclass=Singleton):
        def __new__(cls):
            global instantiated
            instantiated += 1
            return Basic.__new__(cls)

    assert instantiated == 0
    MyNewSingleton()  # force instantiation
    assert instantiated == 1
    assert MyNewSingleton() is not Basic()
    assert MyNewSingleton() is MyNewSingleton()
    assert S.MyNewSingleton is MyNewSingleton()
    assert instantiated == 1

    class MySingleton_sub(MyNewSingleton):
        pass
    assert instantiated == 1
    MySingleton_sub()
    assert instantiated == 2
    assert MySingleton_sub() is not MyNewSingleton()
    assert MySingleton_sub() is MySingleton_sub()


def test_preorder_traversal():
    expr = Basic(b21, b3)
    assert list(
        preorder_traversal(expr)) == [expr, b21, b2, b1, b1, b3, b2, b1]
    assert list(preorder_traversal(('abc', ('d', 'ef')))) == [
        ('abc', ('d', 'ef')), 'abc', ('d', 'ef'), 'd', 'ef']

    result = []
    pt = preorder_traversal(expr)
    for i in pt:
        result.append(i)
        if i == b2:
            pt.skip()
    assert result == [expr, b21, b2, b1, b3, b2]

    expr = z + w*(x + y)
    assert list(preorder_traversal([expr], keys=default_sort_key)) == \
        [[w*(x + y) + z], w*(x + y) + z, z, w*(x + y), w, x + y, x, y]
    assert list(preorder_traversal((x + y)*z, keys=True)) == \
        [z*(x + y), z, x + y, x, y]


def test_sorted_args():
    assert b21._sorted_args == b21.args
    pytest.raises(AttributeError, lambda: x._sorted_args)


def test_call():
    # See the long history of this in issues sympy/sympy#5026 and sympy/sympy#5105.

    pytest.raises(TypeError, lambda: sin(x)({x: 1, sin(x): 2}))
    pytest.raises(TypeError, lambda: sin(x)(1))

    # No effect as there are no callables
    assert sin(x).rcall(1) == sin(x)
    assert (1 + sin(x)).rcall(1) == 1 + sin(x)

    # Effect in the pressence of callables
    l = Lambda(x, 2*x)
    assert (l + x).rcall(y) == 2*y + x
    assert (x**l).rcall(2) == x**4
    # TODO UndefinedFunction does not subclass Expr
    # f = Function('f')
    # assert (2*f)(x) == 2*f(x)


def test_literal_evalf_is_number_is_zero_is_comparable():
    f = Function('f')

    # the following should not be changed without a lot of dicussion
    # `foo.is_number` should be equivalent to `not foo.free_symbols`
    # it should not attempt anything fancy; see is_zero, is_constant
    # and equals for more rigorous tests.
    assert f(1).is_number is True
    i = Integral(0, (x, x, x))
    # expressions that are symbolically 0 can be difficult to prove
    # so in case there is some easy way to know if something is 0
    # it should appear in the is_zero property for that object;
    # if is_zero is true evalf should always be able to compute that
    # zero
    assert i.n() == 0
    assert i.is_zero
    assert i.is_number is False
    assert i.evalf(2, strict=False) == 0

    # issue sympy/sympy#10272
    n = sin(1)**2 + cos(1)**2 - 1
    assert n.is_comparable is not True
    assert n.n(2, strict=False).is_comparable is not True
